"""Anjani telegram utils"""
# Copyright (C) 2020 - 2023  UserbotIndo Team, <https://github.com/userbotindo.git>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import asyncio
import codecs
import html
import re
from enum import IntEnum, unique
from functools import wraps
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncGenerator,
    Callable,
    List,
    Optional,
    Set,
    Tuple,
    Union,
    overload,
)

from pyrogram.client import Client
from pyrogram.enums.chat_member_status import ChatMemberStatus
from pyrogram.enums.chat_members_filter import ChatMembersFilter
from pyrogram.errors import (
    ChannelPrivate,
    ChatForbidden,
    ChatWriteForbidden,
    MessageDeleteForbidden,
    UserNotParticipant,
)
from pyrogram.types import (
    Chat,
    ChatMember,
    InlineKeyboardButton,
    InlineKeyboardMarkup,
    Message,
    User,
)
from typing_extensions import ParamSpecArgs, ParamSpecKwargs

from anjani.util import types as _types
from anjani.util.async_helper import run_sync

if TYPE_CHECKING:
    from anjani.core import Anjani

MESSAGE_CHAR_LIMIT = 4096
STAFF: Set[int] = set()
TRUNCATION_SUFFIX = "... (truncated)"

Button = Union[Tuple[Tuple[str, str, bool]], List[Tuple[str, str, bool]]]


@unique
class Types(IntEnum):
    """A Class representing message type"""

    TEXT = 0
    BUTTON_TEXT = 1
    DOCUMENT = 2
    PHOTO = 3
    VIDEO = 4
    STICKER = 5
    AUDIO = 6
    VOICE = 7
    VIDEO_NOTE = 8
    ANIMATION = 9


def build_button(buttons: Button) -> InlineKeyboardMarkup:
    """Build saved button format"""
    keyb = []  # type: List[List[InlineKeyboardButton]]
    for btn in buttons:
        if btn[2] and keyb:
            keyb[-1].append(InlineKeyboardButton(btn[0], url=btn[1]))
        else:
            keyb.append([InlineKeyboardButton(btn[0], url=btn[1])])
    return InlineKeyboardMarkup(keyb)


def revert_button(button: Button) -> str:
    """Revert button format"""
    res = ""
    for btn in button:
        if btn[2]:
            res += f"\n[{btn[0]}](buttonurl://{btn[1]}:same)"
        else:
            res += f"\n[{btn[0]}](buttonurl://{btn[1]})"
    return res


def parse_button(text: str) -> Tuple[str, Button]:
    """Parse button to save"""
    regex = re.compile(r"(\[([^\[]+?)\]\(buttonurl:(?:/{0,2})(.+?)(:same)?\))")

    prev = 0
    parser_data = ""
    buttons = []  # type: List[Tuple[str, str, bool]]
    for match in regex.finditer(text):
        # escape check
        md_escaped = 0
        to_check = match.start(1) - 1
        while to_check > 0 and text[to_check] == "\\":
            md_escaped += 1
            to_check -= 1

        # if != "escaped" -> Create button: btn
        if md_escaped % 2 == 0:
            # create a thruple with button label, url, and newline status
            buttons.append((match.group(2), match.group(3), bool(match.group(4))))
            parser_data += text[prev : match.start(1)]
            prev = match.end(1)
        # if odd, escaped -> move along
        else:
            parser_data += text[prev:to_check]
            prev = match.start(1) - 1

    parser_data += text[prev:]
    # Remove any markdown button left over if any
    # t = parser_data.rstrip().split()
    # if t:
    #     pattern = re.compile(r"[_-`*~]+")
    #     anyMarkdownLeft = pattern.search(t[-1])
    #     if anyMarkdownLeft:
    #         toRemove = anyMarkdownLeft[0][0]
    #         t[-1] = t[-1].replace(toRemove, "")
    #         return " ".join(t), buttons

    return parser_data.rstrip(), buttons


def get_message_info(msg: Message) -> Tuple[str, Types, Optional[str], Button]:
    """Parse received message and return its content."""
    types = None
    content = None
    text = ""
    buttons = []

    reply_msg = msg.reply_to_message

    if reply_msg:
        text = reply_msg.text or reply_msg.caption
        added_text = None
        if text:
            text, buttons = parse_button(text.markdown)
        else:
            # added_text are from user input
            added_text, buttons = parse_button(msg.text.markdown.split(" ", 2)[-1])

        if not text and added_text is not None:
            text = added_text

        if reply_msg.text:
            types = Types.BUTTON_TEXT if buttons else Types.TEXT
        elif reply_msg.sticker:
            content, types = reply_msg.sticker.file_id, Types.STICKER
        elif reply_msg.document:
            content, types = reply_msg.document.file_id, Types.DOCUMENT
        elif reply_msg.photo:
            content, types = reply_msg.photo.file_id, Types.PHOTO
        elif reply_msg.audio:
            content, types = reply_msg.audio.file_id, Types.AUDIO
        elif reply_msg.voice:
            content, types = reply_msg.voice.file_id, Types.VOICE
        elif reply_msg.video:
            content, types = reply_msg.video.file_id, Types.VIDEO
        elif reply_msg.video_note:
            content, types = reply_msg.video_note.file_id, Types.VIDEO_NOTE
        elif reply_msg.animation:
            content, types = reply_msg.animation.file_id, Types.ANIMATION
        else:
            raise ValueError("Can't get message information")
    else:
        text, buttons = parse_button(msg.text.markdown.split(" ", 2)[2])
        types = Types.BUTTON_TEXT if buttons else Types.TEXT

    return text, types, content, buttons


def truncate(text: str) -> str:
    """Truncates the given text to fit in one Telegram message."""

    if len(text) > MESSAGE_CHAR_LIMIT:
        return text[: MESSAGE_CHAR_LIMIT - len(TRUNCATION_SUFFIX)] + TRUNCATION_SUFFIX

    return text


def is_staff_or_admin(target: ChatMember) -> bool:
    return (
        target.status in {ChatMemberStatus.ADMINISTRATOR, ChatMemberStatus.OWNER}
        or target.user.id in STAFF
    )


def is_staff(target_id: int) -> bool:
    return target_id in STAFF


def mention(user: User) -> str:
    pattern = re.compile(r"<[a-z/][\s\S]*>")
    link = "[{name}](tg://user?id={id})"
    return (
        link.format(name=html.escape(user.first_name), id=user.id)
        if pattern.search(user.first_name)
        else link.format(name=user.first_name, id=user.id)
    )


# { Permission
# Aliases
Bot = ChatMember
Member = ChatMember


async def fetch_permissions(
    client: Client, chat: int, user: int
) -> Tuple[Optional[Bot], Optional[Member]]:
    try:
        bot, member = await asyncio.gather(
            client.get_chat_member(chat, "me"), client.get_chat_member(chat, user)
        )
        return bot, member
    except UserNotParticipant:
        return None, None


# }


# { ChatAdmin
async def get_chat_admins(
    client: Client, chat: int, *, exclude_bot: bool = False
) -> AsyncGenerator[ChatMember, None]:
    member: ChatMember
    async for member in client.get_chat_members(chat, filter=ChatMembersFilter.ADMINISTRATORS):  # type: ignore
        if member.status in {ChatMemberStatus.ADMINISTRATOR, ChatMemberStatus.OWNER}:
            if exclude_bot and member.user.is_bot:
                continue
            yield member


# }


# { Non-Context reply then delete
async def reply_and_delete(message: Message, text: str, del_in: int = 1) -> None:
    if del_in < 1:
        raise ValueError("Delay must be greater than 0")

    try:
        to_del, _ = await asyncio.gather(
            message.reply(text, quote=True),
            asyncio.sleep(del_in),
        )
    except (ChatForbidden, ChannelPrivate, ChatWriteForbidden):
        return

    try:
        await asyncio.gather(message.delete(), to_del.delete())
    except MessageDeleteForbidden:
        pass

    return


# }


# { GetText Language
def __loop_safe(
    func: Callable[
        [
            _types.Bot,
            _types.ChatId,
            _types.TextName,
            ParamSpecArgs,
            _types.NoFormat,
            ParamSpecKwargs,
        ],
        str,
    ]
):  # Special: let default typing choose the return type
    """Decorator for get_text functions"""

    @wraps(func)
    async def wrapper(
        bot: "Anjani",
        chat_id: Optional[int],
        text_name: str,
        *args: Any,
        noformat: bool = False,
        **kwargs: Any,
    ) -> str:
        """Parse the string with user language setting.

        Parameters:
            bot (`Anjani`):
                The bot instance.
            chat_id (`int`, *Optional*):
                Id of the sender(PM's) or chat_id to fetch the user language setting.
                If chat_id is None, the language will always use 'en'.
            text_name (`str`):
                String name to parse. The string is parsed from YAML documents.
            *args (`any`, *Optional*):
                One or more values that should be formatted and inserted in the string.
                The value should be in order based on the language string placeholder.
            noformat (`bool`, *Optional*):
                If True, the text returned will not be formated.
                Default to False.
            **kwargs (`any`, *Optional*):
                One or more keyword values that should be formatted and inserted in the string.
                based on the keyword on the language strings.
        """
        return await run_sync(func, bot, chat_id, text_name, *args, noformat=noformat, **kwargs)

    return wrapper


@__loop_safe
def get_text(
    bot: "Anjani",
    chat_id: Optional[int],
    text_name: str,
    *args: Any,
    noformat: bool = False,
    **kwargs: Any,
) -> str:
    def _get_text(lang: str) -> str:
        try:
            text = codecs.decode(
                codecs.encode(bot.languages[lang][text_name], "latin-1", "backslashreplace"),
                "unicode-escape",
            )
        except KeyError:
            if lang == "en":
                return (
                    f"**NO LANGUAGE STRING FOR '{text_name}' in '{lang}'**\n"
                    "__Please forward this to__ @userbotindo"
                )

            bot.log.warning("NO LANGUAGE STRING FOR '%s' in '%s'", text_name, lang)
            return _get_text("en")
        else:
            try:
                return text if noformat else text.format(*args, **kwargs)
            except (IndexError, KeyError):
                bot.log.error("Failed to format '%s' string on '%s'", text_name, lang)
                raise

    return _get_text(bot.chats_languages.get(chat_id or 0, "en"))


# }


@overload
def get_username(user: Union[User, Chat]) -> Optional[str]:
    ...


@overload
def get_username(user: Union[User, Chat], full: bool = True) -> List[str]:
    ...


def get_username(user: Union[User, Chat], full: bool = False) -> Union[Optional[str], List[str]]:
    """Get username from user"""

    if user.usernames:  # Collectible usernames
        return [i.username for i in user.usernames] if full else user.usernames[0].username

    if user.username:
        return [user.username] if full else user.username

    return [] if full else None



import inspect
import os
import sqlite3
import time
from pathlib import Path
from typing import Any, List, Optional, Tuple, Union

from pyrogram.raw.types.input_peer_channel import InputPeerChannel
from pyrogram.raw.types.input_peer_chat import InputPeerChat
from pyrogram.raw.types.input_peer_user import InputPeerUser
from pyrogram.storage.sqlite_storage import get_input_peer
from pyrogram.storage.storage import Storage

# language=SQLite
SCHEMA = """
CREATE TABLE sessions
(
    dc_id     INTEGER PRIMARY KEY,
    api_id    INTEGER,
    test_mode INTEGER,
    auth_key  BLOB,
    date      INTEGER NOT NULL,
    user_id   INTEGER,
    is_bot    INTEGER
);

CREATE TABLE peers
(
    id             INTEGER PRIMARY KEY,
    access_hash    INTEGER,
    type           INTEGER NOT NULL,
    username       TEXT,
    phone_number   TEXT,
    last_update_on INTEGER NOT NULL DEFAULT (CAST(STRFTIME('%s', 'now') AS INTEGER))
);

CREATE TABLE version
(
    number INTEGER PRIMARY KEY
);

CREATE INDEX idx_peers_id ON peers (id);
CREATE INDEX idx_peers_username ON peers (username);
CREATE INDEX idx_peers_phone_number ON peers (phone_number);

CREATE TRIGGER trg_peers_last_update_on AFTER UPDATE ON peers
BEGIN
    UPDATE peers
    SET last_update_on = CAST(STRFTIME('%s', 'now') AS INTEGER)
    WHERE id = NEW.id;
END;

CREATE TRIGGER trg_session_last_update_on AFTER INSERT ON peers
BEGIN
    UPDATE sessions
    SET date = CAST(STRFTIME('%s', 'now') AS INTEGER);
END;
"""


class SQLiteStorage(Storage):
    VERSION = 3
    USERNAME_TTL = 8 * 60 * 60
    _conn: sqlite3.Connection

    def __init__(self, name: str):
        super().__init__(name)
        self.database = Path(os.getcwd()) / f"anjani/{name}.session"

    async def delete(self):
        raise NotImplementedError

    async def create(self):
        with self.conn:
            self.conn.executescript(SCHEMA)

            self.conn.execute(
                "INSERT INTO sessions VALUES (?, ?, ?, ?, ?, ?, ?)",
                (2, None, None, None, 0, None, None),
            )

            self.conn.execute("INSERT INTO version VALUES (?)", (self.VERSION,))

    async def open(self):
        path = self.database
        file_exists = path.is_file()

        self.conn = sqlite3.connect(database=str(path), timeout=1, check_same_thread=False)

        if not file_exists:
            await self.create()

        self.conn.execute("VACUUM")

    async def save(self):
        await self.date(int(time.time()))

    async def close(self):
        self.conn.close()

    async def update_peers(self, peers: List[Tuple[int, int, str, str, str]]) -> None:
        with self.conn:
            self.conn.executemany(
                "REPLACE INTO peers (id, access_hash, type, username, phone_number)"
                "VALUES (?, ?, ?, ?, ?)",
                peers,
            )

    async def get_peer_by_id(
        self, peer_id: int
    ) -> Union[InputPeerUser, InputPeerChat, InputPeerChannel]:
        r = self.conn.execute(
            "SELECT id, access_hash, type FROM peers WHERE id = ?", (peer_id,)
        ).fetchone()

        if r is None:
            raise KeyError(f"ID not found: {peer_id}")

        return get_input_peer(*r)

    async def get_peer_by_username(
        self, username: str
    ) -> Union[InputPeerUser, InputPeerChat, InputPeerChannel]:
        r = self.conn.execute(
            "SELECT id, access_hash, type, last_update_on FROM peers WHERE username = ?"
            "ORDER BY last_update_on DESC",
            (username,),
        ).fetchone()

        if r is None:
            raise KeyError(f"Username not found: {username}")

        if abs(time.time() - r[3]) > self.USERNAME_TTL:
            raise KeyError(f"Username expired: {username}")

        return get_input_peer(*r[:3])

    async def get_peer_by_phone_number(
        self, phone_number: str
    ) -> Union[InputPeerUser, InputPeerChat, InputPeerChannel]:
        q = self.conn.execute(
            "SELECT id, access_hash, type FROM peers WHERE phone_number = ?", (phone_number,)
        )
        r = q.fetchone()

        if r is None:
            raise KeyError(f"Phone number not found: {phone_number}")

        return get_input_peer(*r)

    async def _get(self) -> Optional[Any]:
        attr = inspect.stack()[2].function

        q = self.conn.execute(f"SELECT {attr} FROM sessions")
        return (q.fetchone())[0]

    async def _set(self, value: Any) -> None:
        attr = inspect.stack()[2].function
        with self.conn:
            self.conn.execute(f"UPDATE sessions SET {attr} = ?", (value,))

    async def _accessor(self, value: Any = object) -> Any:
        return await self._get() if value == object else await self._set(value)

    async def dc_id(self, value=object) -> Optional[int]:
        return await self._accessor(value)

    async def api_id(self, value=object) -> Optional[int]:
        return await self._accessor(value)

    async def test_mode(self, value=object) -> Optional[bool]:
        return await self._accessor(value)

    async def auth_key(self, value=object) -> Optional[bytes]:
        return await self._accessor(value)

    async def date(self, value=object) -> Optional[int]:
        return await self._accessor(value)

    async def user_id(self, value=object) -> Optional[int]:
        return await self._accessor(value)

    async def is_bot(self, value=object) -> Optional[bool]:
        return await self._accessor(value)

    async def version(self, value=object):
        if value == object:
            q = self.conn.execute("SELECT number FROM version")
            return (q.fetchone())[0]
        else:
            with self.conn:
                self.conn.execute("UPDATE version SET number = ?", (value,))



"""Bot Greetings"""
# Copyright (C) 2020 - 2023  UserbotIndo Team, <https://github.com/userbotindo.git>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import asyncio
from html import escape
from typing import (
    Any,
    Callable,
    ClassVar,
    Coroutine,
    List,
    MutableMapping,
    Optional,
    Tuple,
    Union,
)

from pyrogram.client import Client
from pyrogram.enums.parse_mode import ParseMode
from pyrogram.errors import (
    ChannelPrivate,
    ChatWriteForbidden,
    MediaEmpty,
    MessageDeleteForbidden,
    MessageEmpty,
)
from pyrogram.types import Chat, Message, User
from pyrogram.types.messages_and_media.message import Str

from anjani import command, filters, plugin, util
from anjani.util.tg import (
    Button,
    Types,
    build_button,
    get_message_info,
    parse_button,
    revert_button,
)


class Greeting(plugin.Plugin):
    name: ClassVar[str] = "Greetings"
    helpable: ClassVar[bool] = True

    db: util.db.AsyncCollection
    chat_db: util.db.AsyncCollection
    SEND: MutableMapping[int, Callable[..., Coroutine[Any, Any, Optional[Message]]]]

    async def on_load(self) -> None:
        self.db = self.bot.db.get_collection("WELCOME")
        self.chat_db = self.bot.db.get_collection("CHATS")

        self.SEND = {
            Types.TEXT.value: self.bot.client.send_message,
            Types.BUTTON_TEXT.value: self.bot.client.send_message,
            Types.DOCUMENT.value: self.bot.client.send_document,
            Types.PHOTO.value: self.bot.client.send_photo,
            Types.VIDEO.value: self.bot.client.send_video,
            Types.STICKER.value: self.bot.client.send_sticker,
            Types.AUDIO.value: self.bot.client.send_audio,
            Types.VOICE.value: self.bot.client.send_voice,
            Types.VIDEO_NOTE.value: self.bot.client.send_video_note,
            Types.ANIMATION.value: self.bot.client.send_animation,
        }

    async def on_chat_action(self, message: Message) -> None:
        chat = message.chat
        reply_to = message.id
        if message.left_chat_member and message.left_chat_member.id == self.bot.uid:
            return

        # Clean service both for left member and new member if active
        if await self.clean_service(chat.id):
            try:
                await message.delete()
            except (MessageDeleteForbidden, ChannelPrivate):
                pass
            reply_to = 0

        thread_id = await self.get_action_topic(chat)
        if message.chat.is_forum and not thread_id:
            self.log.debug(f"Chat {message.chat.id} is forum but no action topic set!")
            # continue try to send on default (general) topic

        if message.new_chat_members:
            return await self._member_join(message, reply_to, thread_id)

        if message.left_chat_member:
            return await self._member_leave(message, reply_to, thread_id)

    async def _member_leave(
        self, message: Message, reply_to: int, thread_id: Optional[int]
    ) -> None:
        chat = message.chat
        if not await self.is_goodbye(chat.id):
            return

        left_member = message.left_chat_member
        text = await self.left_message(chat.id)
        if not text:
            text = await self.text(chat.id, "default-goodbye", noformat=True)

        formatted_text = await self._build_text(text, left_member, chat, self.bot.client)
        try:
            msg = await self.bot.client.send_message(
                chat.id,
                formatted_text,
                reply_to_message_id=reply_to if not thread_id else None,  # type: ignore
                message_thread_id=thread_id,  # type: ignore
            )
        except ChatWriteForbidden:
            return

        previous = await self.previous_goodbye(chat.id, msg.id)
        if previous:
            try:
                await self.bot.client.delete_messages(chat.id, previous)
            except MessageDeleteForbidden:
                pass

    async def _member_join(self, message: Message, reply_to: int, thread_id: Optional[int]) -> None:
        chat = message.chat
        if not await self.is_welcome(chat.id):
            return

        new_members = message.new_chat_members
        is_bulk_welcome = len(new_members) > 1
        for idx, new_member in enumerate(new_members):
            try:
                if new_member.id == self.bot.uid:
                    await self.bot.client.send_message(
                        chat.id,
                        await self.text(chat.id, "bot-added"),
                        reply_to_message_id=reply_to,
                    )
                else:
                    text, button, msg_type, file_id = await self.welc_message(chat.id)
                    msg_type = Types(msg_type) if msg_type else Types.TEXT
                    if not text:
                        string = await self.text(chat.id, "default-welcome", noformat=True)
                    else:
                        string = text

                    formatted_text = await self._build_text(
                        string, new_member, chat, self.bot.client
                    )

                    if button:
                        button = build_button(button)
                    else:
                        button = None
                    msg = None
                    try:
                        if msg_type in {Types.TEXT, Types.BUTTON_TEXT}:
                            msg = await self.SEND[msg_type](
                                message.chat.id,
                                formatted_text,
                                message_thread_id=thread_id,
                                reply_to_message_id=reply_to,
                                reply_markup=button,
                                disable_web_page_preview=True,
                            )
                        elif msg_type in {Types.STICKER, Types.ANIMATION}:
                            msg = await self.SEND[msg_type](
                                message.chat.id,
                                file_id,
                                message_thread_id=thread_id,
                                reply_to_message_id=reply_to,
                            )
                        else:
                            msg = await self.SEND[msg_type](
                                message.chat.id,
                                file_id,
                                caption=formatted_text,
                                message_thread_id=thread_id,
                                reply_to_message_id=reply_to,
                                reply_markup=button,
                            )
                    except MediaEmpty:
                        await self.bot.client.send_message(
                            message.chat.id,
                            await self.text(message.chat.id, "welcome-message-expired"),
                        )
                    except MessageEmpty:
                        self.log.warning("Welcome message empty on %s.", message.chat.id)

                    if msg:
                        previous = await self.previous_welcome(chat.id, msg.id, is_bulk_welcome)
                        if idx == 0 and previous:
                            try:
                                await self.bot.client.delete_messages(chat.id, previous)
                            except MessageDeleteForbidden:
                                pass
            except ChatWriteForbidden:
                pass

    async def on_chat_migrate(self, message: Message) -> None:
        new_chat = message.chat.id
        old_chat = message.migrate_from_chat_id

        await self.db.update_one(
            {"chat_id": old_chat},
            {"$set": {"chat_id": new_chat}},
        )

    async def on_plugin_backup(self, chat_id: int) -> MutableMapping[str, Any]:
        welcome = await self.db.find_one({"chat_id": chat_id}, {"_id": False})
        return {self.name: welcome} if welcome else {}

    async def on_plugin_restore(self, chat_id: int, data: MutableMapping[str, Any]) -> None:
        await self.db.update_one({"chat_id": chat_id}, {"$set": data[self.name]}, upsert=True)

    @staticmethod
    async def _build_text(
        text: str, user: User, chat: Chat, client: Optional[Client] = None
    ) -> str:
        first_name = user.first_name or ""  # Ensure first name is not None
        last_name = user.last_name
        full_name = first_name + last_name if last_name else first_name
        try:
            count = await client.get_chat_members_count(chat.id) if client else "N/A"
        except ChannelPrivate:
            count = "N/A"

        username = util.tg.get_username(user)

        return text.format(
            first=escape(first_name),
            last=escape(last_name) if last_name else "",
            fullname=escape(full_name),
            username=f"@{username}" if username else user.mention,
            mention=user.mention,
            count=count,
            chatname=escape(chat.title),
            id=user.id,
        )

    async def get_action_topic(self, chat: Chat) -> Optional[int]:
        if not chat.is_forum:
            return None
        data = await self.chat_db.find_one({"chat_id": chat.id}, {"action_topic": True})
        return data.get("action_topic") if data else None

    async def is_welcome(self, chat_id: int) -> bool:
        """Get chat welcome setting"""
        active = await self.db.find_one({"chat_id": chat_id}, {"should_welcome": 1})
        return active.get("should_welcome", True) if active else True

    async def is_goodbye(self, chat_id: int) -> bool:
        """Get chat welcome setting"""
        active = await self.db.find_one({"chat_id": chat_id}, {"should_goodbye": 1})
        return active.get("should_goodbye", True) if active else True

    async def welc_message(
        self, chat_id: int
    ) -> Tuple[Optional[str], Optional[Button], Optional[int], Optional[str]]:
        """Get chat welcome string"""
        message = await self.db.find_one({"chat_id": chat_id})
        if message:
            # This checks data for old welcome schema
            # TODO: deprecate old schema on v3
            if "custom_welcome" in message:
                text: str = message["custom_welcome"]
                button: Optional[Button] = message.get("button")
                message_type: Types = Types.TEXT
                await self.db.delete_one({"chat_id": chat_id})
                await self.set_custom_welcome(
                    chat_id=chat_id,
                    text=text,
                    buttons=button,
                    message_type=message_type,
                    content=None,
                )
                self.log.info("Migrated old welcome message on %d to new schema.", chat_id)
                return text, button, message_type, None
            else:
                return (
                    message.get("text"),
                    message.get("button"),
                    message.get("type"),
                    message.get("file_id"),
                )
        return await self.text(chat_id, "default-welcome", noformat=True), None, None, None

    async def left_message(self, chat_id: int) -> str:
        message = await self.db.find_one({"chat_id": chat_id}, {"custom_goodbye": 1})
        return (
            message.get(
                "custom_goodbye", await self.text(chat_id, "default-goodbye", noformat=True)
            )
            if message
            else await self.text(chat_id, "default-goodbye", noformat=True)
        )

    async def clean_service(self, chat_id: int) -> bool:
        """Fetch clean service setting"""
        clean = await self.db.find_one({"chat_id": chat_id}, {"clean_service": 1})
        if clean:
            return clean.get("clean_service", True)

        return False  # Defaults off

    async def set_custom_welcome(
        self,
        chat_id: int,
        text: str,
        message_type: Types,
        buttons: Optional[Button] = None,
        content: Optional[str] = None,
    ) -> None:
        """Set custom welcome"""
        await self.db.update_one(
            {"chat_id": chat_id},
            {
                "$set": {
                    "text": text,
                    "button": buttons,
                    "file_id": content,
                    "type": message_type,
                }
            },
            upsert=True,
        )

    async def set_custom_goodbye(self, chat_id: int, text: str) -> None:
        """Set custom goodbye"""
        await self.db.update_one({"chat_id": chat_id}, {"$set": {"custom_goodbye": text}})

    async def del_custom_welcome(self, chat_id: int) -> None:
        """Delete custom welcome message"""
        await self.db.update_one(
            {"chat_id": chat_id},
            {
                "$unset": {
                    "custom_welcome": "",
                    "text": "",
                    "button": "",
                    "type": "",
                    "file_id": "",
                }
            },
        )

    async def del_custom_goodbye(self, chat_id: int) -> None:
        """Delete custom goodbye message"""
        await self.db.update_one({"chat_id": chat_id}, {"$unset": {"custom_goodbye": ""}})

    async def greeting_setting(self, chat_id: int, key: str, value: bool) -> None:
        """Turn on/off greetings in chats"""
        if not value:
            await self.db.update_one({"chat_id": chat_id}, {"$set": {key: False}}, upsert=True)
        else:
            await self.db.update_one({"chat_id": chat_id}, {"$unset": {key: ""}}, upsert=True)

    async def previous_welcome(
        self, chat_id: int, msg_id: int, is_bulk: bool = False
    ) -> Union[int, List[int], None]:
        """Save latest welcome msg_id and return previous msg_id"""
        operator = "$push" if is_bulk else "$set"
        data = await self.db.find_one_and_update(
            {"chat_id": chat_id}, {operator: {"prev_welc": msg_id}}, upsert=True
        )
        return data.get("prev_welc", None) if data else None

    async def previous_goodbye(self, chat_id: int, msg_id: int) -> Optional[int]:
        data = await self.db.find_one_and_update(
            {"chat_id": chat_id}, {"$set": {"prev_gdby": msg_id}}, upsert=True
        )
        return data.get("prev_gdby", None) if data else None

    @command.filters(filters.admin_only)
    async def cmd_setwelcome(self, ctx: command.Context) -> str:
        """Set chat welcome message"""
        chat = ctx.chat

        if ctx.input:
            if ctx.message.media:
                # TODO: Add support for command in media caption
                return await self.text(chat.id, "unsupported-media-command")
            else:
                welc_text = (
                    Str(ctx.message.text)
                    .init(ctx.msg.entities)
                    .markdown.split(ctx.invoker, 1)[1]
                    .strip()
                )
                welc_text, buttons = parse_button(welc_text)
                types = Types.TEXT
                content = None
                if ctx.msg.reply_to_message:
                    _, types, content, __ = get_message_info(ctx.msg)
        elif ctx.msg.reply_to_message:
            welc_text, types, content, buttons = get_message_info(ctx.msg)
        else:
            return await self.text(chat.id, "greetings-no-input")

        if not welc_text:
            return await self.text(chat.id, "greetings-button-only-error")

        try:  # Try to build a text first to check message validity
            await self._build_text(
                welc_text or "", ctx.author or self.bot.user, chat, self.bot.client
            )
        except (KeyError, ValueError) as e:
            return await self.text(chat.id, "err-msg-format-parsing", err=e)

        ret, _ = await asyncio.gather(
            self.text(chat.id, "cust-welcome-set"),
            self.set_custom_welcome(chat.id, welc_text, types, buttons, content),
        )
        return ret

    @command.filters(filters.admin_only)
    async def cmd_setgoodbye(self, ctx: command.Context) -> str:
        """Set chat goodbye message"""
        chat = ctx.chat
        if ctx.input:
            gby_text = Str(ctx.input).init(ctx.msg.entities[1:])
        elif ctx.msg.reply_to_message:
            gby_text = ctx.msg.reply_to_message.text or ctx.msg.reply_to_message.caption
        else:
            return await self.text(chat.id, "greetings-no-input")

        ret, _ = await asyncio.gather(
            self.text(chat.id, "cust-goodbye-set"), self.set_custom_goodbye(chat.id, gby_text)
        )
        return ret

    @command.filters(filters.admin_only)
    async def cmd_resetwelcome(self, ctx: command.Context) -> str:
        """Reset saved welcome message"""
        chat = ctx.chat

        ret, _ = await asyncio.gather(
            self.text(chat.id, "reset-welcome"), self.del_custom_welcome(chat.id)
        )
        return ret

    @command.filters(filters.admin_only)
    async def cmd_resetgoodbye(self, ctx: command.Context) -> str:
        """Reset saved welcome message"""
        chat = ctx.chat

        ret, _ = await asyncio.gather(
            self.text(chat.id, "reset-goodbye"), self.del_custom_goodbye(chat.id)
        )
        return ret

    @command.filters(filters.admin_only)
    async def cmd_welcome(self, ctx: command.Context) -> Optional[str]:
        """View current welcome message"""
        chat = ctx.chat
        param = ctx.input.lower()
        noformat = param == "noformat"

        enabled = None
        if param in {"yes", "on", "1"}:
            enabled = True
        elif param in {"no", "off", "0"}:
            enabled = False
        elif param and not noformat:
            return await self.text(chat.id, "err-invalid-option")

        if enabled is not None:
            ret, _ = await asyncio.gather(
                self.text(chat.id, "welcome-set", "on" if enabled else "off"),
                self.greeting_setting(chat.id, "should_welcome", enabled),
            )
            return ret

        (
            setting,
            (text, button, msg_type, file_id),
            clean_service,
        ) = await asyncio.gather(
            self.is_welcome(chat.id), self.welc_message(chat.id), self.clean_service(chat.id)
        )

        if text is None:
            text = ""
        else:
            text += "\n\n"

        if noformat:
            parse_mode = ParseMode.DISABLED
            if button:
                text += revert_button(button)
            button = None
        else:
            parse_mode = ParseMode.MARKDOWN
            if button:
                button = build_button(button)

        view_welc = await self.text(chat.id, "view-welcome", setting, clean_service)
        if ctx.chat.is_forum and not await self.get_action_topic(ctx.chat):
            view_welc += "\n\n" + await self.text(
                chat.id,
                "greetings-topic-default",
                f"https://t.me/{self.bot.user.username}?start=help_topic",
            )

        settings_msg = await ctx.respond(view_welc)

        reply_to = settings_msg.id if settings_msg else None
        try:
            response_text = (
                text
                if text
                else (
                    "Empty, custom welcome message haven't set yet."
                    if not setting
                    else "Default:\n\n" + await self.text(chat.id, "default-welcome", noformat=True)
                )
            )
            msg_type = msg_type or Types.TEXT
            if msg_type in {Types.TEXT, Types.BUTTON_TEXT}:
                await self.SEND[msg_type](
                    ctx.chat.id,
                    response_text,
                    reply_to_message_id=reply_to,
                    reply_markup=button,
                    parse_mode=parse_mode,
                    disable_web_page_preview=True,
                )
            elif msg_type in {Types.STICKER, Types.ANIMATION}:
                await self.SEND[msg_type](
                    ctx.chat.id,
                    file_id,
                    reply_to_message_id=reply_to,
                )
            else:
                await self.SEND[msg_type](
                    ctx.chat.id,
                    file_id,
                    caption=text,
                    reply_to_message_id=reply_to,
                    parse_mode=parse_mode,
                    reply_markup=button,
                )
        except MediaEmpty:
            await self.bot.client.send_message(
                ctx.chat.id, await self.text(ctx.chat.id, "welcome-message-expired")
            )
        except MessageEmpty:
            self.log.warning("Welcome message empty on %s.", ctx.chat.id)

    @command.filters(filters.admin_only)
    async def cmd_goodbye(self, ctx: command.Context) -> Optional[str]:
        """View current goodbye message"""
        chat = ctx.chat
        param = ctx.input.lower()
        noformat = param == "noformat"

        enabled = None
        if param in {"yes", "on", "1"}:
            enabled = True
        elif param in {"no", "off", "0"}:
            enabled = False
        elif param and not noformat:
            return await self.text(chat.id, "err-invalid-option")

        if enabled is not None:
            ret, _ = await asyncio.gather(
                self.text(chat.id, "goodbye-set", "on" if enabled else "off"),
                self.greeting_setting(chat.id, "should_goodbye", enabled),
            )
            return ret

        setting, text, clean_service = await asyncio.gather(
            self.is_goodbye(chat.id), self.left_message(chat.id), self.clean_service(chat.id)
        )

        if noformat:
            parse_mode = ParseMode.DISABLED
        else:
            parse_mode = ParseMode.MARKDOWN

        view_gby = await self.text(chat.id, "view-goodbye", setting, clean_service)
        if ctx.chat.is_forum and not await self.get_action_topic(ctx.chat):
            view_gby += "\n\n" + await self.text(
                chat.id,
                "greetings-topic-default",
                f"https://t.me/{self.bot.user.username}?start=help_topic",
            )

        await ctx.respond(view_gby)
        await ctx.respond(
            text,
            mode="reply",
            parse_mode=parse_mode,
            disable_web_page_preview=True,
        )
        return None

    @command.filters(filters.admin_only)
    async def cmd_cleanservice(self, ctx: command.Context, active: Optional[bool] = None) -> str:
        """Clean service message on new members"""
        chat = ctx.chat

        if active is None:
            return await self.text(chat.id, "err-invalid-option")

        ret, _ = await asyncio.gather(
            self.text(chat.id, "clean-serv-set", "on" if active else "off"),
            self.greeting_setting(chat.id, "clean_service", active),
        )
        return ret



"""User data management"""
# Copyright (C) 2020 - 2023  UserbotIndo Team, <https://github.com/userbotindo.git>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import asyncio
import re
from hashlib import md5
from html import escape
from time import time
from typing import Any, ClassVar, List, Mapping, MutableMapping, Optional, Union

from pyrogram.enums.chat_action import ChatAction
from pyrogram.enums.chat_type import ChatType
from pyrogram.enums.parse_mode import ParseMode
from pyrogram.errors import BadRequest, ChannelInvalid, ChannelPrivate, PeerIdInvalid
from pyrogram.types import CallbackQuery, Chat, ChatPreview, Message, User

from anjani import command, listener, plugin, util

try:
    from userbotindo import get_trust
except ImportError:
    from anjani.util.misc import do_nothing as get_trust


class Users(plugin.Plugin):
    name: ClassVar[str] = "Users"

    chats_db: util.db.AsyncCollection
    users_db: util.db.AsyncCollection
    predict_loaded: bool

    async def on_load(self) -> None:
        self.chats_db = self.bot.db.get_collection("CHATS")
        self.users_db = self.bot.db.get_collection("USERS")
        self.predict_loaded = "SpamPredict" in self.bot.plugins

    def hash_id(self, id: int) -> str:
        # skipcq: PTC-W1003
        return md5((str(id) + self.bot.user.username).encode()).hexdigest()  # skipcq: BAN-B324

    async def build_channel_task(self, channel: Chat) -> asyncio.Task[Any]:
        async def _do_nothing() -> None:
            return

        if channel.type == ChatType.CHANNEL:
            data = await self.chats_db.find_one({"chat_id": channel.id})
            content = {"chat_name": channel.title, "type": "channel"}
            if not data or "hash" not in data:
                content["hash"] = self.hash_id(channel.id)
            return self.bot.loop.create_task(
                self.chats_db.update_one({"chat_id": channel.id}, {"$set": content}, upsert=True)
            )

        return self.bot.loop.create_task(_do_nothing())

    async def build_user_task(self, user: User) -> asyncio.Task:
        data = await self.users_db.find_one({"_id": user.id})
        content: MutableMapping[str, Any] = {
            "username": util.tg.get_username(user),
            "last_seen": int(time()),
        }
        if not data or "hash" not in data:
            content["hash"] = self.hash_id(user.id)
        if not data or "chats" not in data:
            content["chats"] = []
        return self.bot.loop.create_task(
            self.users_db.update_one({"_id": user.id}, {"$set": content}, upsert=True)
        )

    async def on_chat_migrate(self, message: Message) -> None:
        new_chat = message.chat.id
        old_chat = message.migrate_from_chat_id

        await asyncio.gather(
            self.users_db.update_many({"chats": old_chat}, {"$push": {"chats": new_chat}}),
            self.users_db.update_many({"chats": old_chat}, {"$pull": {"chats": old_chat}}),
            self.chats_db.update_one({"chat_id": old_chat}, {"$set": {"chat_id": new_chat}}),
        )

    async def on_chat_action(self, message: Message) -> None:
        """Delete user data from chats"""
        if message.new_chat_members:
            return

        chat = message.chat
        user = message.left_chat_member
        if user.id == self.bot.uid:
            await asyncio.gather(
                self.chats_db.update_one({"chat_id": chat.id}, {"$set": {"member": []}}),
                self.users_db.update_many({"chats": chat.id}, {"$pull": {"chats": chat.id}}),
            )
        else:
            await asyncio.gather(
                self.users_db.update_one({"_id": user.id}, {"$pull": {"chats": chat.id}}),
                self.chats_db.update_one({"chat_id": chat.id}, {"$pull": {"member": user.id}}),
            )

    async def on_callback_query(self, query: CallbackQuery) -> None:
        """Hanle user that sent a callback query"""
        user = query.from_user
        set_content = {"username": util.tg.get_username(user), "name": user.first_name}
        user_data = await self.users_db.find_one({"_id": user.id})

        if self.predict_loaded and (not user_data or "hash" not in user_data):
            set_content["hash"] = self.hash_id(user.id)

        await self.users_db.update_one({"_id": user.id}, {"$set": set_content})

    @listener.priority(50)
    async def on_message(self, message: Message) -> None:
        """Incoming message handler."""
        if message.outgoing:
            return

        chat = message.chat
        user = message.from_user

        if not user or not chat:  # sanity check for service
            return

        tasks = []
        set_content = {
            "username": util.tg.get_username(user),
            "name": user.first_name,
            "last_seen": int(time()),
        }
        user_data = await self.users_db.find_one({"_id": user.id})

        if chat.type == ChatType.PRIVATE:
            if self.predict_loaded:
                if ch := message.forward_from_chat:
                    tasks.append(await self.build_channel_task(ch))
                if not user_data or "hash" not in user_data:
                    set_content["hash"] = self.hash_id(user.id)
                if usr := message.forward_from:
                    tasks.append(await self.build_user_task(usr))

            await asyncio.gather(
                self.users_db.update_one({"_id": user.id}, {"$set": set_content}), *tasks
            )
            return

        chat_data = await self.chats_db.find_one({"chat_id": chat.id})
        chat_update = {
            "$set": {
                "chat_name": chat.title,
                "type": chat.type.name.lower(),
                "last_update": int(time()),
            },
            "$addToSet": {"member": user.id},
        }
        if self.predict_loaded:
            if not user_data or "hash" not in user_data:
                set_content["hash"] = self.hash_id(user.id)
            if not chat_data or "hash" not in chat_data:
                chat_update["$set"]["hash"] = self.hash_id(chat.id)
            update = {
                "$set": set_content,
                "$setOnInsert": {"reputation": 0},
                "$addToSet": {"chats": chat.id},
            }
            if ch := message.forward_from_chat:
                tasks.append(await self.build_channel_task(ch))
            if usr := message.forward_from:
                tasks.append(await self.build_user_task(usr))
        else:
            update = {"$set": set_content, "$addToSet": {"chats": chat.id}}

        await asyncio.gather(
            self.users_db.update_one({"_id": user.id}, update, upsert=True),
            self.chats_db.update_one({"chat_id": chat.id}, chat_update, upsert=True),
            *tasks,
        )

    async def _user_info(self, ctx: command.Context, user: User) -> None:
        """User Info"""
        text = f"<b>{'Bot' if user.is_bot else 'User'} Info</b>\n\n"
        text += f"<b>ID:</b> <code>{user.id}</code>\n"
        text += f"<b>DC ID: </b><code>{user.dc_id if user.dc_id else 'N/A'}</code>\n"
        text += f"<b>First Name: </b>{escape(user.first_name) if user.first_name else ''}\n"
        if user.last_name:
            text += f"<b>Last Name: </b><code>{escape(user.last_name)}</code>\n"

        if username := util.tg.get_username(user, full=True):
            if len(username) == 1:
                text += f"<b>Username: </b>@{username[0]}\n"
            else:
                text += "<b>Usernames: </b>\n"
                text += "\n".join(f" - @{u}" for u in username)
                text += "\n"

        self.bot.client.parse_mode = ParseMode.HTML
        text += f"<b>Permanent user link: </b>{user.mention}\n"
        self.bot.client.parse_mode = ParseMode.MARKDOWN  # telegram_bot.py#L108

        text += (
            "<b>Number of profile pics: </b>"
            f"<code>{await self.bot.client.get_chat_photos_count(user.id)}</code>\n"
        )
        if user.status:
            text += f"<b>Last seen: </b> <code>{user.status.value}</code>\n"

        if user.id in self.bot.staff:
            text += "\nThis person is one of my <b>Staff</b>!\n"
        elif user.is_self:
            text += "\nI've seen them in every chats... wait it's me!!\nWow you're stalking me? 😂"

        if user.is_scam:
            text += "<b>\n⚠️Warning this user is flagged as a scammer by Telegram⚠️</b>\n"

        user_db = await self.users_db.find_one({"_id": user.id})
        if user_db and self.predict_loaded:
            if user_db.get("spam", False):
                text += "<b>\n⚠️Warning I flag this user as a spammer⚠️</b>\n"

            text += f"\n<b>Identifier:</b> <code>{user_db.get('hash', 'unknown')}</code>"
            text += f"\n<b>Reputation:</b> <code>{user_db.get('reputation', 0)}</code>"
            trust = get_trust(user_db.get("pred_sample", []))
            if trust:
                text += f"\n<b>Trust:</b> <code>{trust:.2f}</code>"
            else:
                text += "\n<b>Trust:</b> <code>N/A</code>"

        if user.photo:
            async with ctx.action(ChatAction.UPLOAD_PHOTO):
                file = await self.bot.client.download_media(user.photo.big_file_id)
                if not file:
                    await ctx.respond(text, parse_mode=ParseMode.HTML)
                    return

                try:
                    await self.bot.client.send_photo(
                        ctx.chat.id,
                        file,
                        text,
                        reply_to_message_id=ctx.message.id,
                        parse_mode=ParseMode.HTML,
                    )
                except ValueError:  # Wierd case where the photo id exist but have 0 bytes
                    await ctx.respond(text, parse_mode=ParseMode.HTML)
            return

        await ctx.respond(text, parse_mode=ParseMode.HTML)

    async def _old_user_info(self, data: Mapping[str, Any]) -> str:
        text = "<b>Old User Info</b>\n\n"
        text += f"<b>ID</b>: <code>{data['_id']}</code>\n"
        if data.get("name"):
            text += f"<b>Name</b>: {escape(data['name'])}\n"
        if data.get("username"):
            text += f"<b>Username</b>: @{data['username']}\n"
        if self.predict_loaded:
            text += f"\n<b>Identifier</b>: <code>{data.get('hash', 'unknown')}</code>"
            text += f"\n<b>Reputation</b>: <code>{data.get('reputation', 0)}</code>"
        text += f"\nI've seen them on {len(data.get('chats', []))} chats."
        return text

    async def _chat_info(self, ctx: command.Context, chat: Union[Chat, ChatPreview]) -> None:
        """Chat Info"""
        text = "<b>Chat Info</b>\n\n"
        if isinstance(chat, ChatPreview):
            text += f"<b>Chat Type:</b> <code>{chat.type}</code>\n"
            text += f"<b>Title:</b> <code>{escape(chat.title)}</code>\n"
            text += f"<b>Member Count:</b> <code>{chat.members_count}</code>\n"
        else:
            text += f"<b>ID:</b> <code>{chat.id}</code>\n"
            if chat.dc_id:
                text += f"<b>DC ID:</b> <code>{chat.dc_id}</code>\n"
            text += f"<b>Chat Type:</b> <code>{chat.type.__dict__['_name_']}</code>\n"
            text += f"<b>Title:</b> <code>{chat.title}</code>\n"

            if username := util.tg.get_username(chat, full=True):
                if len(username) == 1:
                    text += f"<b>Chat Username: </b>@{username[0]}\n"
                else:
                    text += "<b>Chat Usernames: </b>\n"
                    text += "\n".join(f" - @{u}" for u in username)
                    text += "\n"

            text += f"<b>Member Count:</b> <code>{chat.members_count}</code>\n"
            if chat.linked_chat:
                text += f"<b>Linked Chat:</b> <code>{chat.linked_chat.title}</code>\n"

            if self.predict_loaded:
                chat_data = await self.chats_db.find_one({"chat_id": chat.id})
                if chat_data:
                    text += f"<b>Identifier:</b> <code>{chat_data.get('hash', 'unknown')}</code>\n"

        if chat.photo:
            async with ctx.action(ChatAction.UPLOAD_PHOTO):
                file = await self.bot.client.download_media(chat.photo.big_file_id)  # type: ignore
                if not file:
                    await ctx.respond(text, parse_mode=ParseMode.HTML)
                    return

                await self.bot.client.send_photo(
                    ctx.chat.id,
                    file,
                    text,
                    reply_to_message_id=ctx.message.id,
                    parse_mode=ParseMode.HTML,
                )
            return

        await ctx.respond(text, parse_mode=ParseMode.HTML)
        return

    async def _old_chat_info(self, data: Mapping[str, Any]) -> str:
        text = "<b>Old Chat Info</b>\n\n"
        text += f"<b>ID:</b> <code>{data['chat_id']}</code>\n"
        text += f"<b>Chat Name:</b> {escape(data['chat_name'])}\n"
        if self.predict_loaded:
            text += f"\n<b>Identifier:</b> <code>{data.get('hash', 'unknown')}</code>"
        return text

    async def cmd_info(self, ctx: command.Context, args: Optional[str] = None) -> Optional[str]:
        """Fetch a telegram peer data"""
        if not args:
            if ctx.msg.reply_to_message:
                if ctx.msg.reply_to_message.from_user:
                    return await self._user_info(ctx, ctx.msg.reply_to_message.from_user)
                if ctx.msg.reply_to_message.sender_chat:
                    return await self._chat_info(ctx, ctx.msg.reply_to_message.sender_chat)

            if not ctx.msg.reply_to_message:
                if not ctx.author:
                    return await self.get_text(ctx.chat.id, "err-anonymous")

                return await self._user_info(ctx, ctx.author)

            return

        id_match = None
        if self.predict_loaded:
            id_match = re.search(r"([a-fA-F\d]{32})", args)

        if id_match:
            user_data = await self.users_db.find_one({"hash": id_match.group(0)})
            if user_data:
                try:
                    user = await ctx.bot.client.get_users(user_data["_id"])
                    if isinstance(user, List):
                        user = user[0]

                    return await self._user_info(ctx, user)
                except PeerIdInvalid:
                    text = await self._old_user_info(user_data)
                    if text:
                        await ctx.respond(text, parse_mode=ParseMode.HTML)
                        return

            chat_data = await self.chats_db.find_one({"hash": id_match.group(0)})
            if chat_data:
                try:
                    chat = await ctx.bot.client.get_chat(chat_data["chat_id"])
                    return await self._chat_info(ctx, chat)
                except (PeerIdInvalid, ChannelInvalid, ChannelPrivate):
                    text = await self._old_chat_info(chat_data)
                    if text:
                        await ctx.respond(text, parse_mode=ParseMode.HTML)
                        return

            return await self.text(ctx.chat.id, "err-invalid-pid")

        try:
            user = await ctx.bot.client.get_users(args)
            if isinstance(user, List):
                user = user[0]

            return await self._user_info(ctx, user)
        except (IndexError, BadRequest):  # chat peer
            try:
                uid = int(args)
            except (TypeError, ValueError):
                pass
            else:
                try:
                    chat = await ctx.bot.client.get_chat(uid)
                    return await self._chat_info(ctx, chat)
                except (BadRequest, ChannelPrivate):
                    user = await self.users_db.find_one({"_id": uid})
                    if user:
                        text = await self._old_user_info(user)
                        if text:
                            await ctx.respond(text, parse_mode=ParseMode.HTML)
                            return

                    chat = await self.chats_db.find_one({"chat_id": uid})
                    if chat:
                        text = await self._old_chat_info(chat)
                        if text:
                            await ctx.respond(text, parse_mode=ParseMode.HTML)
                            return
        except KeyError:  # Rare case username expired, so make it recursively
            return await self.cmd_info(ctx, args)

        return await self.text(ctx.chat.id, "err-peer-invalid")



"""staff's commands"""
# Copyright (C) 2020 - 2023  UserbotIndo Team, <https://github.com/userbotindo.git>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import asyncio
from io import BytesIO
from typing import ClassVar, Optional, Set

from aiopath import AsyncPath
from pyrogram.enums.chat_type import ChatType
from pyrogram.errors.exceptions.bad_request_400 import (
    ChannelInvalid,
    PeerIdInvalid,
    UserNotParticipant,
)

from anjani import command, filters, plugin, util


class Staff(plugin.Plugin):
    name: ClassVar[str] = "Staff Tools"

    db: util.db.AsyncCollection

    async def on_load(self) -> None:
        self.db = self.bot.db.get_collection("CHATS")

    @command.filters(filters.owner_only)
    async def cmd_broadcast(self, ctx: command.Context) -> Optional[str]:
        """Broadcast a message to all chats"""
        if not ctx.input:
            return "Give me a message to send."

        await ctx.respond("Sending broadcast...")

        text = ctx.input + "\n\n*This is a broadcast message."
        tasks: Set[asyncio.Task] = set()
        async for chat in self.db.find({}, {"chat_id": 1, "type": 1}):
            if chat.get("type") == "channel":
                continue
            if len(tasks) % 25 == 0:
                # sleep every 25 msg tasks to prevent flood limit.
                await asyncio.sleep(1)

            task = self.bot.loop.create_task(self.bot.client.send_message(chat["chat_id"], text))
            tasks.add(task)

        failed = 0
        sent = 0
        done, _ = await asyncio.wait(tasks)
        for fut in done:
            try:
                fut.result()
            except (PeerIdInvalid, ChannelInvalid):
                failed += 1
            else:
                sent += 1

        return (
            "Broadcast complete!\n"
            f"{sent} groups succeed, {failed} groups failed to receive the message"
        )

    @command.filters(filters.staff_only)
    async def cmd_leavechat(self, ctx: command.Context) -> str:
        """leave the given chat_id"""
        if not ctx.args or not ctx.input:
            return "Give me the chat id!"

        try:
            await self.bot.client.leave_chat(ctx.args[0])
        except (PeerIdInvalid, UserNotParticipant):
            return "I'm not a member on that group"
        else:
            return "I left the group"

    @command.filters(filters.staff_only)
    async def cmd_chatlist(self, ctx: command.Context, get_all: Optional[bool] = False) -> None:
        """Send file of chat's I'm in"""
        chatfile = "List of chats.\n"
        async for chat in self.db.find({}, {"chat_id": 1, "chat_name": 1, "type": 1}):
            if not get_all and chat.get("type") == "channel":
                continue

            name = chat.get("chat_name")
            if not name:
                name = "[NO CHAT NAME]"

            chatfile += f"{name} | ({chat['chat_id']})\n"

        with BytesIO(str.encode(chatfile)) as output:
            output.name = "chatlist.txt"
            await ctx.msg.reply_document(
                document=output,
                caption="Here is the list of chats in my database.",
            )

    @command.filters(filters.dev_only)
    async def cmd_logs(self, ctx: command.Context) -> None:
        """Send bot log"""
        file = AsyncPath("Anjani.log")
        if ctx.message.chat.type != ChatType.PRIVATE:
            await ctx.respond("I've send the log on PM's")

        await self.bot.client.send_document(
            ctx.author.id,
            str(file),
            caption="**Bot Logs**",
            force_document=True,
        )



"""Anjani base telegram"""
# Copyright (C) 2020 - 2023  UserbotIndo Team, <https://github.com/userbotindo.git>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import asyncio
import signal
import sys
from functools import partial
from hashlib import sha256
from typing import TYPE_CHECKING, Any, MutableMapping, Optional, Set, Tuple, Type, Union

import pyrogram.filters as flt
from aiocache import cached
from aiopath import AsyncPath
from pyrogram.client import Client
from pyrogram.enums.parse_mode import ParseMode
from pyrogram.filters import Filter
from pyrogram.handlers.callback_query_handler import CallbackQueryHandler
from pyrogram.handlers.chat_member_updated_handler import ChatMemberUpdatedHandler
from pyrogram.handlers.inline_query_handler import InlineQueryHandler
from pyrogram.handlers.message_handler import MessageHandler
from pyrogram.types import (
    CallbackQuery,
    Chat,
    ChatMember,
    ChatPreview,
    InlineQuery,
    Message,
    User,
)
from yaml import full_load

from anjani import util
from anjani.language import get_lang_file
from anjani.util.cache_limiter import CacheLimiter

from .anjani_mixin_base import MixinBase
from .sqlite_storage import SQLiteStorage

if TYPE_CHECKING:
    from .anjani_bot import Anjani

EventType = Union[CallbackQuery, InlineQuery, Message]
TgEventHandler = Union[
    CallbackQueryHandler, InlineQueryHandler, MessageHandler, ChatMemberUpdatedHandler
]


class TelegramBot(MixinBase):
    # Initialized during instantiation
    __running: bool
    _limiter: CacheLimiter
    _plugin_event_handlers: MutableMapping[str, Tuple[TgEventHandler, int]]

    loaded: bool
    staff: Set[int]
    devs: Set[int]
    chats_languages: MutableMapping[int, str]
    languages: MutableMapping[str, MutableMapping[str, str]]

    # Initialized during startup
    client: Client
    user: User
    uid: int
    start_time_us: int
    owner: int

    def __init__(self: "Anjani", **kwargs: Any) -> None:
        self.__running = False
        self._limiter = CacheLimiter(ttl=10, max_value=3)
        self._plugin_event_handlers = {}

        self.loaded = False
        self.staff = set()
        self.devs = set()
        self.chats_languages = {}
        self.languages = {}

        # Propagate initialization to other mixins
        super().__init__(**kwargs)

    async def init_client(self: "Anjani") -> None:
        api_id = int(self.config.API_ID)
        api_hash = self.config.API_HASH
        bot_token = self.config.BOT_TOKEN

        try:
            self.owner = int(self.config.OWNER_ID)
        except KeyError:
            self.log.warning("Owner id is not set! you won't be able to run staff command!")
            self.owner = 0

        if "--fresh" not in sys.argv:
            # Load session from database
            data = await self.db.get_collection("SESSION").find_one(
                {"_id": sha256(str(bot_token).encode()).hexdigest()}
            )
            file = AsyncPath("anjani/anjani.session")
            if data and not await file.exists():
                self.log.info("Loading session from database")
                await file.write_bytes(data["session"])

        # Initialize Telegram client with gathered parameters
        self.client = Client(
            name="anjani",
            api_id=api_id,
            api_hash=api_hash,
            bot_token=bot_token,
            workdir="anjani",
            workers=self.config.WORKERS,
            parse_mode=ParseMode.MARKDOWN,
            storage=SQLiteStorage("anjani"),
        )

    async def start(self: "Anjani") -> None:
        if self.__running:
            raise RuntimeError("This bot instance is already running")

        self.log.info("Starting")
        await self.init_client()

        # Register core command handler
        self.client.add_handler(MessageHandler(self.on_command, self.command_predicate()), -1)

        # Load plugin
        self.load_all_plugins()
        await self.dispatch_event("load")
        self.loaded = True

        async with asyncio.Lock():
            # Start Telegram client
            try:
                await self.client.start()
            except AttributeError:
                self.log.error(
                    "Unable to get input for authorization! Make sure all configuration are done before running the bot."
                )
                raise

            user = await self.client.get_me()
            if not isinstance(user, User):
                raise TypeError("Missing full self user information")

            self.user = user
            # noinspection PyTypeChecker
            self.uid = user.id

        self.staff.add(self.owner)
        self.devs.add(self.owner)

        # Update staff from db
        async for doc in self.db.get_collection("STAFF").find():
            if doc["rank"] == "dev":
                self.devs.add(doc["_id"])

            self.staff.add(doc["_id"])

        # Update global staff variable
        util.tg.STAFF.update(self.staff)

        # Update Language setting chat from db
        async for data in self.db.get_collection("LANGUAGE").find({}, {"_id": False}):
            self.chats_languages[data["chat_id"]] = data["language"]

        # Load text from language file
        async for language_file in get_lang_file():
            self.languages[language_file.stem] = await util.run_sync(
                full_load, await language_file.read_text()
            )

        # Record start time and dispatch start event
        self.start_time_us = util.time.usec()
        await self.dispatch_event("start", self.start_time_us)

        self.log.info("Bot is ready")

        if not self.config.is_flag_active("disable_catchup"):
            self.log.info("Catching up on missed events")
            await self.dispatch_missed_events()
            self.log.info("Finished catching up")

        # Dispatch final late start event
        await self.dispatch_event("started")

    async def idle(self: "Anjani") -> None:
        if self.__running:
            raise RuntimeError("This bot instance is already running")

        signals = {
            k: v
            for v, k in signal.__dict__.items()
            if v.startswith("SIG") and not v.startswith("SIG_")
        }
        task: asyncio.Task

        if sys.platform == "win32":

            def clear_handler() -> None:
                for signame in (signal.SIGINT, signal.SIGBREAK, signal.SIGABRT):
                    signal.signal(signame, signal.SIG_DFL)

            def signal_handler_windows(signum: int, *args: Any) -> None:
                print(flush=True)
                self.log.info(f"Stop signal received ('{signals[signum]}').")
                self.__running = False
                clear_handler()
                task.cancel()

            for signame in (signal.SIGINT, signal.SIGBREAK, signal.SIGABRT):
                signal.signal(signame, signal_handler_windows)
        else:

            def clear_handler() -> None:
                for signame in (signal.SIGINT, signal.SIGTERM, signal.SIGABRT):
                    self.loop.remove_signal_handler(signame)

            def signal_handler(signum: int) -> None:
                print(flush=True)  # Separate signal and next log
                self.log.info(f"Stop signal received ('{signals[signum]}').")
                self.__running = False
                clear_handler()
                task.cancel()

            for signame in (signal.SIGINT, signal.SIGTERM, signal.SIGABRT):
                self.loop.add_signal_handler(signame, partial(signal_handler, signame))

        self.__running = True
        while self.__running:
            task = self.loop.create_task(asyncio.sleep(150))

            try:
                await task
            except asyncio.CancelledError:
                pass

    async def run(self: "Anjani") -> None:
        if self.__running:
            raise RuntimeError("This bot instance is already running")

        try:
            # Start client
            try:
                await self.start()
            except KeyboardInterrupt:
                self.log.warning("Received interrupt while connecting")
                return

            if self.config.IS_CI:
                self.log.info("Completed CI run, exiting")
                return

            # Request updates, then idle until disconnected
            await self.idle()
        finally:
            # Make sure we stop when done
            await self.stop()

    def update_plugin_event(
        self: "Anjani",
        name: str,
        event_type: Type[TgEventHandler],
        *,
        filters: Optional[Filter] = None,
        group: int = 0,
    ) -> None:
        if name in self.listeners:
            # Add if there ARE listeners and it's NOT already registered
            if name not in self._plugin_event_handlers:

                async def event_handler(
                    client: Client, event: EventType  # skipcq: PYL-W0613
                ) -> None:
                    user = event.from_user
                    if name == "callback_query" and await self._limiter.exceeded(user.id):
                        return await event.answer("")  # type: ignore

                    try:
                        await self.dispatch_event(name, event)
                    finally:
                        if name == "callback_query":
                            return await self._limiter.increment(user.id)

                if filters is not None:
                    handler_info = (event_type(event_handler, filters), group)
                else:
                    handler_info = (event_type(event_handler), group)

                self.client.add_handler(*handler_info)
                self._plugin_event_handlers[name] = handler_info
        elif name in self._plugin_event_handlers:
            # Remove if there are NO listeners and it's ALREADY registered
            self.client.remove_handler(*self._plugin_event_handlers[name])
            del self._plugin_event_handlers[name]

    def update_plugin_events(self: "Anjani") -> None:
        self.update_plugin_event("callback_query", CallbackQueryHandler)
        self.update_plugin_event(
            "chat_action", MessageHandler, filters=flt.new_chat_members | flt.left_chat_member
        )
        self.update_plugin_event("chat_member_update", ChatMemberUpdatedHandler, group=1)
        self.update_plugin_event(
            "chat_migrate", MessageHandler, filters=flt.migrate_from_chat_id, group=1
        )
        self.update_plugin_event("inline_query", InlineQueryHandler, group=1)
        self.update_plugin_event(
            "message",
            MessageHandler,
            filters=~flt.new_chat_members
            & ~flt.left_chat_member
            & ~flt.migrate_from_chat_id
            & ~flt.migrate_to_chat_id,
            group=-1,
        )

    @property
    def events_activated(self: "Anjani") -> int:
        return len(self._plugin_event_handlers)

    def redact_message(self: "Anjani", text: str) -> str:
        api_id = self.config.API_ID
        api_hash = self.config.API_HASH
        bot_token = self.config.BOT_TOKEN
        db_uri = self.config.DB_URI

        if api_id in text:
            text = text.replace(api_id, "[REDACTED]")
        if api_hash in text:
            text = text.replace(api_hash, "[REDACTED]")
        if bot_token in text:
            text = text.replace(bot_token, "[REDACTED]")
        if db_uri in text:
            text = text.replace(db_uri, "[REDACTED]")

        return text

    async def respond(
        self: "Anjani",
        msg: Message,
        text: str = "",
        *,
        mode: Optional[str] = "edit",
        redact: bool = True,
        response: Optional[Message] = None,
        **kwargs: Any,
    ) -> Message:
        async def reply(reference: Message, *, text: str = "", **kwargs: Any) -> Message:
            if animation := kwargs.pop("animation", None):
                return await reference.reply_animation(animation, caption=text, **kwargs)
            if audio := kwargs.pop("audio", None):
                return await reference.reply_audio(audio, caption=text, **kwargs)
            if document := kwargs.pop("document", None):
                return await reference.reply_document(document, caption=text, **kwargs)
            if photo := kwargs.pop("photo", None):
                return await reference.reply_photo(photo, caption=text, **kwargs)
            if video := kwargs.pop("video", None):
                return await reference.reply_video(video, caption=text, **kwargs)

            return await reference.reply(text, **kwargs)

        if text:
            # Redact sensitive information if enabled and known
            if redact:
                text = self.redact_message(text)
            # Truncate messages longer than Telegram's 4096-character length limit
            text = util.tg.truncate(text)

        # get rid of emtpy value "animation", "audio", "document", "photo", "video"
        for key, value in dict(kwargs).items():
            if key in {"animation", "audio", "document", "photo", "video"} and not value:
                del kwargs[key]

        # force reply and as default behaviour if response is None
        if mode == "reply" or response is None and mode == "edit":
            return await reply(msg, text=text, **kwargs)

        # Only accept edit if we already respond the original msg
        if response is not None and mode == "edit":
            if any(key in kwargs for key in ("animation", "audio", "document", "photo", "video")):
                # Make client re-send the message with the new media instead editing a text
                await response.delete()
                return await reply(msg, text=text or response.text, **kwargs)

            if "reply_to_message_id" in kwargs:
                del kwargs["reply_to_message_id"]

            return await response.edit(text, **kwargs)

        raise ValueError(f"Unknown response mode {mode}")

    @cached(ttl=60)
    async def get_chat(self: "Anjani", chat_id: int) -> Union[Chat, ChatPreview]:
        """Wrapper for `Client.get_chat` with a TTL cache."""
        return await self.client.get_chat(chat_id)

    @cached(ttl=60)
    async def get_chat_member(self: "Anjani", chat_id: int, user_id: int) -> ChatMember:
        """Wrapper for `Client.get_chat_member` with a TTL cache."""
        return await self.client.get_chat_member(chat_id, user_id)



"""Lock/Unlock group Plugin"""
# Copyright (C) 2020 - 2023  UserbotIndo Team, <https://github.com/userbotindo.git>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import time
import unicodedata as ud
from collections import OrderedDict
from datetime import datetime
from typing import (
    Any,
    Callable,
    ClassVar,
    Coroutine,
    List,
    MutableMapping,
    Optional,
    Set,
    Union,
)

from pyrogram.client import Client
from pyrogram.enums.chat_member_status import ChatMemberStatus
from pyrogram.enums.chat_type import ChatType
from pyrogram.enums.message_entity_type import MessageEntityType
from pyrogram.errors import (
    ChannelPrivate,
    ChatAdminRequired,
    MessageDeleteForbidden,
    MessageIdInvalid,
    PeerIdInvalid,
    UserNotParticipant,
)
from pyrogram.types import Chat, ChatPermissions, Message

from anjani import command, filters, plugin, util


async def anon(_: Client, message: Message) -> bool:
    return bool(message.sender_chat)


async def button(_: Client, message: Message) -> bool:
    return bool(message.reply_markup)


async def rtl(_: Client, message: Message) -> bool:
    text = message.text or message.caption
    if not text:
        return False

    checkers = await Lockings.detect_alphabet(text)
    return "arabic" in checkers or "hebrew" in checkers


async def url(_: Client, message: Message) -> bool:
    if not message.entities:
        return False

    for entity in message.entities:
        if entity.type == MessageEntityType.URL:
            return True

    return False


LOCK_TYPES = OrderedDict(
    sorted(
        {
            "audio": filters.audio,
            "animation": filters.animation,
            "document": filters.document,
            "forward": filters.forwarded,
            "photo": filters.photo,
            "sticker": filters.sticker,
            "video": filters.video,
            "contact": filters.contact,
            "location": filters.location,
            "venue": filters.venue,
            "game": filters.game,
            "dice": filters.dice,
            "button": button,
            "inline": filters.via_bot,
            "url": url,
            "bots": "bots",
            "rtl": rtl,
            "anon": anon,
        }.items()
    )
)


class Lockings(plugin.Plugin):
    name: ClassVar[str] = "Lockings"
    helpable: ClassVar[bool] = True

    db: util.db.AsyncCollection
    restrictions: MutableMapping[str, MutableMapping[str, MutableMapping[str, bool]]]

    async def on_load(self) -> None:
        self.db = self.bot.db.get_collection("LOCKINGS")
        self.restrictions = {
            "lock": self.get_restrictions("lock"),
            "unlock": self.get_restrictions("unlock"),
        }

    async def on_chat_migrate(self, message: Message) -> None:
        new_chat = message.chat.id
        old_chat = message.migrate_from_chat_id

        await self.db.update_one(
            {"chat_id": old_chat},
            {"$set": {"chat_id": new_chat}},
        )

    async def on_plugin_backup(self, chat_id: int) -> MutableMapping[str, Any]:
        data = await self.db.find_one({"chat_id": chat_id}, {"_id": False})
        return {self.name: data} if data else {}

    async def on_plugin_restore(self, chat_id: int, data: MutableMapping[str, Any]) -> None:
        await self.db.update_one({"chat_id": chat_id}, {"$set": data[self.name]}, upsert=True)

    async def on_message(self, message: Message) -> None:
        if message.outgoing:
            return

        chat = message.chat
        user = message.from_user
        if not chat or chat.type not in (ChatType.GROUP, ChatType.SUPERGROUP):
            return

        if not user and message.sender_chat:
            if message.sender_chat.id == chat.id:  # anon admin
                return

            current_chat: Chat = await self.bot.get_chat(chat.id)
            if current_chat.linked_chat and message.sender_chat.id == current_chat.linked_chat.id:
                # Linked channel group
                return

        if user:
            try:
                target = await chat.get_member(user.id)
            except (ChatAdminRequired, ChannelPrivate, PeerIdInvalid, UserNotParticipant):
                pass
            else:
                if target.status in (ChatMemberStatus.ADMINISTRATOR, ChatMemberStatus.OWNER):
                    return

        locked = await self.get_chat_restrictions(chat.id)
        for lock_type in locked:
            try:
                func: Union[
                    str, Callable[[Client, Message], Coroutine[Any, Any, bool]]
                ] = LOCK_TYPES[lock_type]
                if not callable(func):
                    continue

                if await func(self.bot.client, message):
                    await message.delete()
            except MessageDeleteForbidden:
                await self.bot.respond(
                    message,
                    await self.get_text(chat.id, "lockings-failed-to-delete", lock_type=lock_type),
                    quote=True,
                )
            except MessageIdInvalid:
                continue
            except Exception as e:  # skipcq: PYL-W0703
                self.log.error(e, exc_info=e)
                continue

    async def on_chat_action(self, action: Message) -> None:
        chat = action.chat
        added_by = action.from_user
        if action.left_chat_member:
            return

        locked = await self.get_chat_restrictions(chat.id)
        if not locked or locked and "bots" not in locked:
            return

        bot_perm, added_by_perm = await util.tg.fetch_permissions(
            self.bot.client, chat.id, added_by.id
        )
        if not (bot_perm and added_by_perm) or added_by_perm.status == ChatMemberStatus.OWNER:
            return  # bot added by owner

        # Kick the bot if it's not added by the owner
        for member in action.new_chat_members:
            if not member.is_bot:
                continue

            if bot_perm.status != ChatMemberStatus.ADMINISTRATOR:
                await self.bot.respond(
                    action, await self.get_text(chat.id, "lockings-bots-not-admin"), quote=True
                )
                return  # Tell once so we don't spam the chat

            try:
                await chat.ban_member(
                    member.id, until_date=datetime.fromtimestamp(int(time.time() + 30))
                )
            except UserNotParticipant:  # Bot probably kicked already
                continue

    @staticmethod
    async def detect_alphabet(ustring: str) -> Set[str]:
        return {ud.name(char).split(" ")[0].lower() for char in ustring if char.isalpha()}

    @staticmethod
    def get_mode(mode: str) -> bool:
        return mode != "lock"

    def get_restrictions(self, mode: str) -> MutableMapping[str, MutableMapping[str, bool]]:
        permission = self.get_mode(mode)
        return OrderedDict(
            sorted(
                {
                    "all": {
                        "can_send_messages": permission,
                        "can_send_media_messages": permission,
                        "can_send_polls": permission,
                        "can_send_other_messages": permission,
                        "can_add_web_page_previews": permission,
                        "can_change_info": permission,
                        "can_invite_users": permission,
                        "can_pin_messages": permission,
                        "can_send_audios": permission,
                        "can_send_docs": permission,
                        "can_send_games": permission,
                        "can_send_gifs": permission,
                        "can_send_inline": permission,
                        "can_send_photos": permission,
                        "can_send_plain": permission,
                        "can_send_roundvideos": permission,
                        "can_send_stickers": permission,
                        "can_send_videos": permission,
                        "can_send_voice": permission,
                    },
                    "messages": {"can_send_messages": permission},
                    "media": {"can_send_media_messages": permission},
                    "polls": {"can_send_polls": permission},
                    "previews": {"can_add_web_page_previews": permission},
                    "info": {"can_change_info": permission},
                    "invite": {"can_invite_users": permission},
                    "pin": {"can_pin_messages": permission},
                    "audio": {"can_send_audios": permission},
                    "docs": {"can_send_docs": permission},
                    "games": {"can_send_games": permission},
                    "gifs": {"can_send_gifs": permission},
                    "inline": {"can_send_inline": permission},
                    "photos": {"can_send_photos": permission},
                    "plain": {"can_send_plain": permission},
                    "videonote": {"can_send_roundvideos": permission},
                    "stickers": {"can_send_stickers": permission},
                    "videos": {"can_send_videos": permission},
                    "voices": {"can_send_voice": permission},
                }.items()
            )
        )

    async def get_chat_restrictions(self, chat_id: int) -> List[str]:
        data = await self.db.find_one({"chat_id": chat_id}, {"type": 1})
        return data["type"] if data else []

    def unpack_permissions(
        self, permissions: MutableMapping[str, bool], mode: str, lock_type: str
    ) -> ChatPermissions:
        try:
            del permissions["_client"]
        except KeyError:
            pass

        permissions.update(self.restrictions[mode][lock_type])
        return ChatPermissions(**permissions)

    async def update_chat_restrictions(self, chat_id: int, types: str, mode: str) -> None:
        if mode == "lock":
            aggregation = "$push"
        elif mode == "unlock":
            aggregation = "$pull"
        else:
            raise ValueError("Invalid mode")

        await self.db.update_one({"chat_id": chat_id}, {aggregation: {"type": types}}, upsert=True)

    @command.filters(filters.admin_only, aliases={"listlocks", "locks", "locked", "locklist"})
    async def cmd_list_locks(self, ctx: command.Context) -> str:
        chat = ctx.chat
        permissions = dict(chat.permissions.__dict__)
        try:
            del permissions["_client"]
        except KeyError:
            pass

        text = await ctx.get_text("lockings-types-perm")
        for types, is_lock in permissions.items():
            if not is_lock:
                text += f"\n × `{types}`"

        locked = await self.get_chat_restrictions(chat.id)
        if not locked and all(value is True for value in permissions.values()):
            return await ctx.get_text("lockings-types-list-empty")

        text += "\n\n" + await ctx.get_text("lockings-types-list")
        for types in sorted(locked):
            text += f"\n × `{types}`"

        return text

    @command.filters(filters.admin_only)
    async def cmd_lock(self, ctx: command.Context, lock_type: Optional[str] = None) -> str:
        chat = ctx.chat
        if not lock_type:
            return await ctx.get_text("lockings-type-required")

        lock_type = lock_type.lower()
        permissions = dict(chat.permissions.__dict__)
        if lock_type in self.restrictions["lock"]:
            if (
                not permissions[list(self.restrictions["lock"][lock_type].keys())[0]]
                and lock_type != "all"
            ):
                return await ctx.get_text("lockings-type-locked", lock_type=lock_type)

            try:
                await self.bot.client.set_chat_permissions(
                    chat_id=chat.id,
                    permissions=self.unpack_permissions(permissions, "lock", lock_type),
                )
            except ChatAdminRequired as e:
                return await ctx.get_text(
                    "lockings-admin-required",
                    message=e.MESSAGE.split()[-1].strip(")").split(".")[-1].strip('"')
                    if e.MESSAGE
                    else "",
                )
        elif lock_type in LOCK_TYPES:
            locked = await self.get_chat_restrictions(chat.id)
            if lock_type in locked:
                return await ctx.get_text("lockings-type-locked", lock_type=lock_type)

            await self.update_chat_restrictions(chat.id, lock_type, "lock")
        else:
            return await ctx.get_text("lockings-type-invalid", lock_type=lock_type)

        return await ctx.get_text("lockings-type-done", lock_type=lock_type)

    @command.filters(filters.admin_only, aliases={"locktypes"})
    async def cmd_lock_types(self, ctx: command.Context) -> str:
        text = ""
        for types in sorted(list(LOCK_TYPES) + list(self.restrictions["lock"])):
            text += f"\n × `{types}`"

        return await ctx.get_text("lockings-types-available") + text

    @command.filters(filters.admin_only)
    async def cmd_unlock(self, ctx: command.Context, unlock_type: Optional[str] = None) -> str:
        chat = ctx.chat
        if not unlock_type:
            return await ctx.get_text("unlockings-type-required")

        unlock_type = unlock_type.lower()
        permissions = dict(chat.permissions.__dict__)
        if unlock_type in self.restrictions["unlock"]:
            if (
                permissions[list(self.restrictions["unlock"][unlock_type].keys())[0]]
                and unlock_type != "all"
            ):
                return await ctx.get_text("unlockings-type-unlocked", unlock_type=unlock_type)

            await self.bot.client.set_chat_permissions(
                chat_id=chat.id,
                permissions=self.unpack_permissions(permissions, "unlock", unlock_type),
            )
        elif unlock_type in LOCK_TYPES:
            locked = await self.get_chat_restrictions(chat.id)
            if not locked or unlock_type not in locked:
                return await ctx.get_text("unlockings-type-unlocked", unlock_type=unlock_type)

            await self.update_chat_restrictions(chat.id, unlock_type, "unlock")
        else:
            return await ctx.get_text("unlockings-type-invalid", unlock_type=unlock_type)

        return await ctx.get_text("unlockings-type-done", unlock_type=unlock_type)



"""Member muting plugin"""
# Copyright (C) 2020 - 2023  UserbotIndo Team, <https://github.com/userbotindo.git>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import asyncio
from datetime import datetime
from typing import ClassVar, Optional

from pyrogram.errors import PeerIdInvalid, UsernameInvalid, UsernameNotOccupied
from pyrogram.errors.exceptions.bad_request_400 import UserNotParticipant
from pyrogram.types import ChatMember, ChatPermissions, Message, User

from anjani import command, filters, plugin, util


class Muting(plugin.Plugin):
    name: ClassVar[str] = "Muting"
    helpable: ClassVar[bool] = True

    async def _muter(self, message: Message, user: User, time: int = 0, flag: str = "") -> str:
        chat_id = message.chat.id
        user_id = user.id
        mstring = "mute-success-time" if time else "mute-success"
        try:
            await self.bot.client.restrict_chat_member(
                chat_id, user_id, ChatPermissions(), datetime.fromtimestamp(time)
            )

            return await self.text(chat_id, mstring, user.first_name, flag)
        except (UsernameInvalid, UsernameNotOccupied, PeerIdInvalid):
            return await self.text(chat_id, "err-invalid-username-id")

    @command.filters(filters.can_restrict)
    async def cmd_mute(
        self, ctx: command.Context, member: Optional[ChatMember] = None, flag: str = ""
    ) -> str:
        """Mute Chat Member"""
        chat_id = ctx.chat.id
        user = None

        if not member:
            if ctx.args and not ctx.args[0].endswith(("s", "m", "h")):
                return await self.text(chat_id, "no-mute-user")

            if ctx.msg.reply_to_message and ctx.msg.reply_to_message.from_user:
                try:
                    member = await self.bot.client.get_chat_member(
                        chat_id, ctx.msg.reply_to_message.from_user.id
                    )
                except UserNotParticipant:
                    # User is not a participant in the chat (replying from channel discussion)
                    user = await self.bot.client.get_users(ctx.msg.reply_to_message.from_user.id)
                    if isinstance(user, list):
                        user = user[0]

                flag = ctx.args[0] if ctx.args else ""
            else:
                return await self.text(chat_id, "no-mute-user")

        if member is not None:
            user = member.user
            if util.tg.is_staff_or_admin(member):
                return await self.text(chat_id, "cant-mute-admin")

            if member.permissions and not member.permissions.can_send_messages:
                return await self.text(chat_id, "already-muted")

        if not user:
            return await self.text(chat_id, "no-mute-user")

        if user.id == self.bot.uid:
            return await self.text(chat_id, "self-muting")

        if util.tg.is_staff(user.id):
            return await self.text(chat_id, "cant-mute-admin")

        if flag:
            until = util.time.extract_time(flag)
            if not until:
                return await self.text(chat_id, "invalid-time-flag")
        else:
            until = 0

        return await self._muter(ctx.msg, user, until, flag)

    @command.filters(filters.can_restrict)
    async def cmd_unmute(self, ctx: command.Context, member: Optional[ChatMember] = None) -> str:
        """Unmute chat member"""
        chat_id = ctx.chat.id
        if member is None:
            if ctx.args:
                return await self.text(chat_id, "err-peer-invalid")

            if ctx.msg.reply_to_message and ctx.msg.reply_to_message.from_user:
                member = await self.bot.client.get_chat_member(
                    chat_id, ctx.msg.reply_to_message.from_user.id
                )
            else:
                return await self.text(chat_id, "no-unmute-user")

        if member.permissions and not member.permissions.can_send_messages:
            _, t = await asyncio.gather(
                ctx.message.chat.unban_member(member.user.id),
                self.text(chat_id, "unmute-done")
            )
            return t

        return await self.text(chat_id, "user-not-muted")



""" Admin reporting plugin """
# Copyright (C) 2020 - 2023  UserbotIndo Team, <https://github.com/userbotindo.git>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import asyncio
from typing import Any, MutableMapping, Optional

from pyrogram.enums.chat_member_status import ChatMemberStatus
from pyrogram.enums.chat_type import ChatType
from pyrogram.errors import UserNotParticipant
from pyrogram.types import Message

from anjani import command, filters, listener, plugin, util


class Reporting(plugin.Plugin):
    name = "Reporting"
    helpable = True

    db: util.db.AsyncCollection
    user_db: util.db.AsyncCollection

    async def on_load(self) -> None:
        self.db = self.bot.db.get_collection("CHAT_REPORTING")
        self.user_db = self.bot.db.get_collection("USER_REPORTING")

    async def on_chat_migrate(self, message: Message) -> None:
        new_chat = message.chat.id
        old_chat = message.migrate_from_chat_id

        await self.db.update_one(
            {"chat_id": old_chat},
            {"$set": {"chat_id": new_chat}},
        )

    async def on_plugin_backup(self, chat_id: int) -> MutableMapping[str, Any]:
        report = await self.db.find_one({"chat_id": chat_id}, {"_id": False})
        return {self.name: report} if report else {}

    async def on_plugin_restore(self, chat_id: int, data: MutableMapping[str, Any]) -> None:
        await self.db.update_one({"chat_id": chat_id}, {"$set": data[self.name]}, upsert=True)

    @listener.filters(filters.regex(r"(?i)^@admin(s)?\b") & filters.group & ~filters.outgoing)
    async def on_message(self, message: Message) -> None:
        chat = message.chat
        user = message.from_user
        if not await self.is_active(chat.id, is_private=False):
            return

        # Anonymous, so ignores it
        if not user:
            return

        try:
            invoker = await chat.get_member(user.id)
            if invoker.status in {ChatMemberStatus.OWNER, ChatMemberStatus.ADMINISTRATOR}:
                return  # ignore command from admins
        except UserNotParticipant:
            pass  # keep going when user is not a member

        if not message.reply_to_message:
            await message.reply(await self.text(chat.id, "no-report-user"))
            return

        reported_user = message.reply_to_message.from_user
        if not reported_user:
            return

        if reported_user.id == self.bot.uid:
            await message.reply_text(await self.text(chat.id, "cant-report-me"))
            return
        if reported_user.id == message.from_user.id:
            await message.reply_text(await self.text(chat.id, "cant-self-report"))
            return

        try:
            member = await chat.get_member(reported_user.id)
        except UserNotParticipant:
            await message.reply_text(await self.text(chat.id, "user-not-in-chat"))
            return

        if util.tg.is_staff_or_admin(member):
            await message.reply_text(await self.text(chat.id, "cant-report-admin"))
            return

        reply_text = await self.text(chat.id, "report-notif", reported_user.mention)
        slots = 4096 - len(reply_text)
        async for admin in util.tg.get_chat_admins(self.bot.client, chat.id, exclude_bot=True):
            if await self.is_active(admin.user.id, True):
                reply_text += f"[\u200b](tg://user?id={admin.user.id})"

            slots -= 1
            if slots == 0:
                break

        await message.reply_text(reply_text)

    async def setting(self, chat_id: int, is_private: bool, setting: bool) -> None:
        if is_private:
            if setting:
                await self.user_db.update_one(
                    {"_id": chat_id}, {"$set": {"setting": True}}, upsert=True
                )
            else:
                await self.user_db.delete_one({"_id": chat_id})
        else:
            if setting:
                await self.db.update_one(
                    {"chat_id": chat_id}, {"$set": {"setting": True}}, upsert=True
                )
            else:
                await self.db.delete_one({"chat_id": chat_id})

    async def is_active(self, uid: int, is_private: bool) -> bool:
        """Get current setting default to True"""
        if is_private:
            data = await self.user_db.find_one({"_id": uid})
        else:
            data = await self.db.find_one({"chat_id": uid})
        if not data:
            return True

        return data.get("setting", True)

    @command.filters(filters.group)
    async def cmd_report(self, ctx: command.Context) -> None:
        return await self.on_message(ctx.message)

    @command.filters(filters.admin_only)
    async def cmd_reports(
        self, ctx: command.Context, setting: Optional[bool] = None
    ) -> Optional[str]:
        """Report setting command"""
        if ctx.msg.reply_to_message:
            return None

        chat = ctx.chat
        private = chat.type == ChatType.PRIVATE

        if setting is None:
            if not ctx.input:
                return await self.text(
                    chat.id,
                    "report-setting" if private else "chat-report-setting",
                    await self.is_active(chat.id, private),
                )

            return await self.text(chat.id, "err-yes-no-args")

        _, member = await util.tg.fetch_permissions(self.bot.client, chat.id, ctx.author.id)
        if not member or member.status not in {
            ChatMemberStatus.ADMINISTRATOR,
            ChatMemberStatus.OWNER,
        }:
            return None

        if setting is True:
            text = "report-on" if private else "chat-report-on"
        else:
            text = "report-off" if private else "chat-report-off"

        ret, _ = await asyncio.gather(
            self.text(chat.id, text),
            self.setting(chat.id, private, setting),
        )
        return ret



"""Anjani example-plugin"""
# Copyright (C) 2020 - 2023  UserbotIndo Team, <https://github.com/userbotindo.git>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import asyncio
import io
import sys
from typing import IO, Any, ClassVar, MutableMapping

if sys.version_info >= (3, 10):
    from aiopath import AsyncPurePosixPath as PosixPath
else:
    from aiopath import PureAsyncPosixPath as PosixPath

from pyrogram.types import CallbackQuery, Message

from anjani import command, filters, listener, plugin, util


class ExamplePlugin(plugin.Plugin):
    name: ClassVar[str] = "Example"
    disabled: ClassVar[bool] = True
    helpable: ClassVar[bool] = False

    db: util.db.AsyncCollection

    async def on_load(self) -> None:
        self.db = self.bot.db.get_collection("example")

    # Filter Listener only on group chats
    @listener.filters(filters.group)
    async def on_message(self, message: Message) -> None:
        self.log.info(f"Received message: {message.text}")
        await self.db.update_one({"_id": message.id}, {"$set": {"text": message.text}}, upsert=True)

    async def on_callback_query(self, query: CallbackQuery) -> None:
        self.log.info("Button clicked: %s", query.data)
        await query.answer("You clicked the button!")

    async def on_chat_action(self, message: Message) -> None:
        if message.new_chat_members:
            for new_member in message.new_chat_members:
                self.log.info("New member joined: %s", new_member.first_name)
        else:
            left_member = message.left_chat_member
            self.log.info("A member just left chat: %s", left_member.first_name)

    async def on_chat_migrate(self, message: Message) -> None:
        self.log.info("Migrating chat...")
        new_chat = message.chat.id
        old_chat = message.migrate_from_chat_id

        await self.db.update_one(
            {"chat_id": old_chat},
            {"$set": {"chat_id": new_chat}},
        )

    async def on_plugin_backup(self, chat_id: int) -> MutableMapping[str, Any]:
        """Dispatched when /backup command is Called"""
        self.log.info("Backing up data plugin: %s", self.name)
        data = await self.db.find_one({"chat_id": chat_id}, {"_id": False})
        if not data:
            return {}

        return {self.name: data}

    async def on_plugin_restore(self, chat_id: int, data: MutableMapping[str, Any]) -> None:
        """Dispatched when /restore command is Called"""
        self.log.info("Restoring data plugin: %s", self.name)
        await self.db.update_one({"chat_id": chat_id}, {"$set": data[self.name]}, upsert=True)

    async def cmd_test(self, ctx: command.Context) -> str:
        await ctx.respond("Processing...")
        await asyncio.sleep(1)

        if ctx.input:
            return ctx.input

        return "It works!"

    async def get_cat(self) -> IO[bytes]:
        # Get the link to a random cat picture
        async with self.bot.http.get("https://aws.random.cat/meow") as resp:
            # Read and parse the response as JSON
            json = await resp.json()
            # Get the "file" field from the parsed JSON object
            cat_url = json["file"]

        # Get the actual cat picture
        async with self.bot.http.get(cat_url) as resp:
            # Get the data as a byte array (bytes object)
            cat_data = await resp.read()

        # Construct a byte stream from the data.
        # This is necessary because the bytes object is immutable, but we need to add a "name" attribute to set the
        # filename. This facilitates the setting of said attribute without altering behavior.
        cat_stream = io.BytesIO(cat_data)

        # Set the name of the cat picture before sending.
        # This is necessary for Pyrogram to detect the file type and send it as a photo/GIF rather than just a plain
        # unnamed file that doesn't render as media in clients.
        # We abuse aiopath to extract the filename section here for convenience, since URLs are *mostly* POSIX paths
        # with the exception of the protocol part, which we don't care about here.
        cat_stream.name = PosixPath(cat_url).name

        return cat_stream

    # Filter Command only on private
    # And create alias command /get_cat
    @command.filters(filters.private, aliases=["get_cat"])
    async def cmd_cat(self, ctx: command.Context) -> None:
        await ctx.respond("Fetching cat...")
        cat_stream = await self.get_cat()

        await ctx.respond(photo=cat_stream)



"""Bot Federation Tools"""
# Copyright (C) 2020 - 2023  UserbotIndo Team, <https://github.com/userbotindo.git>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import asyncio
from datetime import datetime
from typing import (
    Any,
    AsyncIterator,
    Dict,
    List,
    Mapping,
    MutableMapping,
    Optional,
    Union,
)
from uuid import uuid4

from aiopath import AsyncPath
from pyrogram.enums.chat_member_status import ChatMemberStatus
from pyrogram.enums.chat_type import ChatType
from pyrogram.errors import (
    BadRequest,
    ChannelPrivate,
    ChatAdminRequired,
    Forbidden,
    PeerIdInvalid,
    UserAdminInvalid,
)
from pyrogram.types import (
    CallbackQuery,
    Chat,
    ChatMemberUpdated,
    InlineKeyboardButton,
    InlineKeyboardMarkup,
    Message,
    User,
)

from anjani import command, filters, listener, plugin, util


class Federation(plugin.Plugin):
    name = "Federations"
    helpable = True

    db: util.db.AsyncCollection

    async def on_load(self) -> None:
        self.db = self.bot.db.get_collection("FEDERATIONS")
        self.chat_db = self.bot.db.get_collection("CHATS")

    async def on_chat_migrate(self, message: Message) -> None:
        new_chat = message.chat.id
        old_chat = message.migrate_from_chat_id

        await self.db.update_one(
            {"chat_id": old_chat},
            {"$set": {"chat_id": new_chat}},
        )

    async def on_chat_action(self, message: Message) -> None:
        if message.left_chat_member:
            return

        chat = message.chat
        if not chat:
            return
        fed_data = await self.get_fed_bychat(chat.id)
        if not fed_data:
            return

        if message.new_chat_members:
            for new_member in message.new_chat_members:
                banned = await self.is_fbanned(chat.id, new_member.id)
                if banned:
                    await self.fban_handler(chat, new_member, banned)

        if message.left_chat_member and message.left_chat_member.id == self.bot.uid:
            fed_data = await self.get_fed_bychat(chat.id)
            if fed_data:
                # Leave the chat federation
                await self.db.update_one({"_id": fed_data["_id"]}, {"$pull": {"chats": chat.id}})

    async def on_chat_member_update(self, update: ChatMemberUpdated) -> None:
        """Leave federation if bot is demoted"""
        if not (update.old_chat_member and update.new_chat_member):
            return
        if update.old_chat_member.user.id != self.bot.uid:
            return

        old_priv = update.old_chat_member.privileges
        new_priv = update.new_chat_member.privileges

        if (old_priv and old_priv.can_restrict_members) and not (
            new_priv and new_priv.can_restrict_members
        ):
            chat = update.chat
            fed_data = await self.get_fed_bychat(chat.id)
            if not fed_data:
                return
            ret, _ = await asyncio.gather(
                self.text(chat.id, "fed-autoleave", fed_data["name"], fed_data["_id"]),
                self.db.update_one({"_id": fed_data["_id"]}, {"$pull": {"chats": chat.id}}),
            )
            thread_id = await self.get_action_topic(chat.id)
            await self.bot.client.send_message(
                chat.id,
                ret,
                message_thread_id=thread_id,  # type: ignore
            )

    @listener.filters(filters.regex(r"(rm|log)fed_(.*)"))
    async def on_callback_query(self, query: CallbackQuery) -> Any:
        """Delete federation button listener"""
        cmd = query.matches[0].group(1)
        arg = query.matches[0].group(2)
        chat = query.message.chat
        user = query.from_user

        if cmd == "rm":
            if arg == "abort":
                return await query.message.edit_text(
                    await self.text(chat.id, "fed-delete-canceled")
                )

            data = await self.db.find_one_and_delete({"_id": arg})
            await query.message.edit_text(await self.text(chat.id, "fed-delete-done", data["name"]))
        elif cmd == "log":
            owner_id, fid = arg.split("_")
            if user.id != int(owner_id):
                return await query.edit_message_text(
                    await self.text(chat.id, "fed-invalid-identity")
                )

            data = await self.db.find_one_and_update({"_id": fid}, {"$set": {"log": chat.id}})
            await query.edit_message_text(
                await self.text(chat.id, "fed-log-set-chnl", data["name"])
            )
        else:
            raise ValueError("Invalid callback data command")

    async def on_message(self, message: Message) -> None:
        if message.outgoing or not message.chat:
            return

        chat = message.chat
        if chat.type == ChatType.CHANNEL and message.text and message.text.startswith("/setfedlog"):
            return await self.channel_setlog(message)

        target = message.from_user or message.sender_chat
        if not target:
            return

        banned = await self.is_fbanned(chat.id, target.id)
        if banned:
            await self.fban_handler(chat, target, banned)

    async def get_action_topic(self, chat_id: int) -> Optional[int]:
        data = await self.chat_db.find_one({"chat_id": chat_id}, {"action_topic": 1})
        return data.get("action_topic") if data else None

    @staticmethod
    def is_fed_admin(data: Mapping[str, Any], user: int) -> bool:
        """Check federation admin"""
        return user == data["owner"] or user in data.get("admins", [])

    async def get_fed_bychat(self, chat: int) -> Optional[Mapping[str, Any]]:
        return await self.db.find_one({"chats": chat})

    async def get_fed_byowner(self, user: int) -> Optional[Mapping[str, Any]]:
        return await self.db.find_one({"owner": user})

    async def get_fed(self, fid: str) -> Optional[Mapping[str, Any]]:
        return await self.db.find_one({"_id": fid})

    async def _get_fed_subs_str(self, fid: str) -> Optional[str]:
        """Get federation that subcribe current federation as string"""
        res = ""
        async for i in self.db.find({"subscribers": fid}):
            res += f"- **{i['name']}** (`{i['_id']}`)\n"
        return res or None

    async def _get_fed_subs_data(self, fid: str) -> AsyncIterator[Mapping[str, Any]]:
        """Get federation that subcribe current federation"""
        async for i in self.db.find(
            {"subscribers": fid}, {"_id": 1, "name": 1, "banned": 1, "banned_chat": 1}
        ):
            yield i

    async def fban_user(
        self,
        fid: str,
        user: int,
        *,
        fullname: Optional[str] = None,
        reason: Optional[str] = None,
    ) -> None:
        """Fban a user"""
        await self.db.update_one(
            {"_id": fid},
            {
                "$set": {
                    f"banned.{user}": {"name": fullname, "reason": reason, "time": datetime.now()}
                }
            },
            upsert=True,
        )

    async def fban_chat(
        self,
        fid: str,
        chat: int,
        *,
        title: Optional[str] = None,
        reason: Optional[str] = None,
    ) -> None:
        """Fban a channel"""
        await self.db.update_one(
            {"_id": fid},
            {
                "$set": {
                    f"banned_chat.{chat}": {
                        "title": title,
                        "reason": reason,
                        "time": datetime.now(),
                    }
                }
            },
        )

    async def unfban_user(self, fid: str, user: int) -> None:
        """Remove banned user"""
        await self.db.update_one({"_id": fid}, {"$unset": {f"banned.{user}": None}}, upsert=True)

    async def unfban_chat(self, fid: str, chat: int) -> None:
        """Remove banned chat"""
        await self.db.update_one(
            {"_id": fid}, {"$unset": {f"banned_chat.{chat}": None}}, upsert=True
        )

    async def check_fban(self, target: int) -> util.db.AsyncCursor:
        """Check user banned list"""
        return self.db.find(
            {
                "$or": [
                    {f"banned.{target}": {"$exists": True}},
                    {f"banned_chat.{target}": {"$exists": True}},
                ]
            },
            {
                "_id": 1,
                "name": 1,
                f"banned.{target}": 1,
                f"banned_chat.{target}": 1,
            },
        )

    async def is_fbanned(self, chat: int, target: int) -> Optional[MutableMapping[str, Any]]:
        def check(target: int, data: Mapping[str, Any]) -> Optional[MutableMapping[str, Any]]:
            """Scoped check to verify if user is banned"""
            if str(target) in data.get("banned", {}):
                user_data = data["banned"][str(target)]
                user_data["fed_name"] = data["name"]
                user_data["type"] = "user"
                return user_data

            if str(target) in data.get("banned_chat", {}):
                channel_data = data["banned_chat"][str(target)]
                channel_data["fed_name"] = data["name"]
                channel_data["type"] = "chat"
                return channel_data

        data = await self.db.find_one(
            {
                "chats": chat,
                "$or": [
                    {f"banned.{target}": {"$exists": True}},
                    {f"banned_chat.{target}": {"$exists": True}},
                ],
            },
            {
                f"banned.{target}": 1,
                f"banned_chat.{target}": 1,
                "name": 1,
            },
        )
        if data:
            res = check(target, data)
            if res:
                return res

        # Check if user is banned in subcribed federation
        fid = await self.get_fed_bychat(chat)
        if not fid:
            return None

        subscribe = [res async for res in self._get_fed_subs_data(fid["_id"])]
        if not subscribe:
            return None
        for i in subscribe:
            res = check(target, i)
            if res:
                res["subfed"] = True
                return res

    async def fban_handler(
        self, chat: Chat, user: Union[User, Chat], data: MutableMapping[str, Any]
    ) -> None:
        string = "fed-autoban"
        if data["type"] != "user":
            string += "-chat"
        if data.get("subfed", False):
            string += "-subfed"

        try:
            await asyncio.gather(
                self.bot.client.send_message(
                    chat.id,
                    await self.text(
                        chat.id,
                        string,
                        user.mention
                        if (data["type"] == "user" and isinstance(user, User))
                        else data["title"],
                        data["fed_name"],
                        data["reason"],
                        data["time"].strftime("%Y %b %d %H:%M UTC"),
                    ),
                    message_thread_id=await self.get_action_topic(chat.id),  # type: ignore
                ),
                self.bot.client.ban_chat_member(chat.id, user.id),
            )
        except (ChatAdminRequired, UserAdminInvalid) as e:
            self.log.warning(
                f"Can't ban user {util.tg.get_username(user)} on {chat.title} ({chat.id}) caused by {e}"
            )

    async def cmd_newfed(self, ctx: command.Context, name: Optional[str] = None) -> str:
        """Create a new federations"""
        chat = ctx.chat
        if chat.type != ChatType.PRIVATE:
            return await self.text(chat.id, "err-chat-private")

        if not name:
            return await self.text(chat.id, "need-fed-name")

        fed_id = str(uuid4())
        owner = ctx.msg.from_user

        exists = await self.db.find_one({"owner": owner.id})
        if exists:
            return await self.text(chat.id, "federation-limit")

        await self.db.insert_one({"_id": fed_id, "name": name, "owner": owner.id, "log": owner.id})
        return await self.text(chat.id, "new-federation", fed_name=name, fed_id=fed_id)

    async def cmd_delfed(self, ctx: command.Context) -> Optional[str]:
        """Delete federations"""
        chat = ctx.chat
        if chat.type != ChatType.PRIVATE:
            return await self.text(chat.id, "err-chat-private")

        owner = ctx.msg.from_user

        exists = await self.db.find_one({"owner": owner.id})
        if not exists:
            return await self.text(chat.id, "user-no-feds")

        await ctx.respond(
            await self.text(chat.id, "del-fed-confirm", exists["name"]),
            reply_markup=InlineKeyboardMarkup(
                [
                    [
                        InlineKeyboardButton(
                            text=await self.text(chat.id, "fed-confirm-text"),
                            callback_data="rmfed_" + exists["_id"],
                        )
                    ],
                    [
                        InlineKeyboardButton(
                            text=await self.text(chat.id, "fed-abort-text"),
                            callback_data="rmfed_abort",
                        )
                    ],
                ]
            ),
        )
        return None

    @command.filters(filters.admin_only & filters.can_restrict)
    async def cmd_joinfed(self, ctx: command.Context, fid: Optional[str] = None) -> str:
        """Join a federation in chats"""
        chat = ctx.chat
        user = ctx.msg.from_user

        invoker = await chat.get_member(user.id)
        if invoker.status != ChatMemberStatus.OWNER:
            return await self.text(chat.id, "err-group-creator-cmd")

        if not fid:
            return await self.text(chat.id, "fed-not-found")

        if await self.get_fed_bychat(chat.id):
            return await self.text(chat.id, "fed-cant-two-feds")

        data = await self.get_fed(fid)
        if not data:
            return await self.text(chat.id, "fed-invalid-id")

        if chat.id in data.get("chats", []):
            return await self.text(chat.id, "fed-already-connected")

        ret, _ = await asyncio.gather(
            self.text(chat.id, "fed-chat-joined-info", data["name"]),
            self.db.update_one({"_id": fid}, {"$push": {"chats": chat.id}}),
        )
        if log := data.get("log"):
            await self.bot.client.send_message(
                log,
                f"**New Chat Joined Federation**\n**Name**: {chat.title}",
            )

        return ret

    @command.filters(filters.admin_only)
    async def cmd_leavefed(self, ctx: command.Context) -> str:
        """Leave a federation in chats"""
        chat = ctx.chat
        user = ctx.msg.from_user

        invoker = await chat.get_member(user.id)
        if invoker.status != ChatMemberStatus.OWNER:
            return await self.text(chat.id, "err-group-creator-cmd")

        fed = await self.get_fed_bychat(chat.id)

        if not fed:
            return await self.text(chat.id, "fed-not-connected")

        ret, _ = await asyncio.gather(
            self.text(chat.id, "fed-chat-leave-info", fed["name"]),
            self.db.update_one({"_id": fed["_id"]}, {"$pull": {"chats": chat.id}}),
        )

        if log := fed.get("log"):
            await self.bot.client.send_message(
                log,
                f"**New Chat Joined Federation**\n**Name**: {chat.title}",
            )
        return ret

    @command.filters(aliases=["fpromote"])
    async def cmd_fedpromote(self, ctx: command.Context, user: Optional[User] = None) -> str:
        """Promote user to fed admin"""
        chat = ctx.chat
        if chat.type == ChatType.PRIVATE:
            return await self.text(chat.id, "err-chat-groups")

        if not user:
            if not ctx.msg.reply_to_message:
                return await self.text(chat.id, "fed-no-promote-user")
            user = ctx.msg.reply_to_message.from_user

        data = await self.get_fed_bychat(chat.id)
        if not data:
            return await self.text(chat.id, "fed-no-fed-chat")

        if ctx.author.id != data["owner"]:
            return await self.text(chat.id, "fed-owner-only-promote")
        if user.id == data["owner"]:
            return await self.text(chat.id, "fed-already-owner")
        if user.id in data.get("admins", []):
            return await self.text(chat.id, "fed-already-admin")

        ret, _ = await asyncio.gather(
            self.text(chat.id, "fed-promote-done"),
            self.db.update_one({"_id": data["_id"]}, {"$push": {"admins": user.id}}),
        )
        if log := data.get("log"):
            await self.bot.client.send_message(
                log,
                "**New Fed Promotion**\n"
                "**Fed**: " + data["name"] + "\n"
                f"**Promoted FedAdmin**: {user.mention}\n"
                f"**User ID**: `{user.id}`",
            )

        return ret

    @command.filters(aliases=["fdemote"])
    async def cmd_feddemote(self, ctx: command.Context, user: Optional[User] = None) -> str:
        """Demote user to fed admin"""
        chat = ctx.chat
        if chat.type == ChatType.PRIVATE:
            return await self.text(chat.id, "err-chat-groups")

        invoker = ctx.msg.from_user
        if not user:
            if not ctx.msg.reply_to_message:
                return await self.text(chat.id, "fed-no-demote-user")
            user = ctx.msg.reply_to_message.from_user

        data = await self.get_fed_bychat(chat.id)
        if not data:
            return await self.text(chat.id, "fed-no-fed-chat")

        if invoker.id != data["owner"]:
            return await self.text(chat.id, "fed-owner-only-demote")
        if user.id == data["owner"]:
            return await self.text(chat.id, "fed-already-owner")
        if user.id not in data.get("admins", []):
            return await self.text(chat.id, "fed-user-not-admin")

        ret, _ = await asyncio.gather(
            self.text(chat.id, "fed-demote-done"),
            self.db.update_one({"_id": data["_id"]}, {"$pull": {"admins": user.id}}),
        )
        if log := data.get("log"):
            await self.bot.client.send_message(
                log,
                "**New Fed Demotion**\n"
                "**Fed**: " + data["name"] + "\n"
                f"**Promoted FedAdmin**: {user.mention}\n"
                f"**User ID**: `{user.id}`",
            )

        return ret

    async def cmd_fedinfo(self, ctx: command.Context, fid: Optional[str] = None) -> str:
        """Fetch federation info"""
        chat = ctx.chat

        if fid is not None:
            data = await self.get_fed(fid)
            if not data:
                return await self.text(chat.id, "fed-invalid-id")
        elif chat.type != ChatType.PRIVATE:
            data = await self.get_fed_bychat(chat.id)
            if not data:
                return (
                    await self.text(chat.id, "fed-no-fed-chat")
                    + "\n"
                    + await self.text(chat.id, "fed-specified-id")
                )
        else:
            return await self.text(chat.id, "fed-specified-id")

        owner = await self.bot.client.get_users(data["owner"])
        if isinstance(owner, List):
            owner = owner[0]

        res = await self.text(
            chat.id,
            "fed-info-text",
            data["_id"],
            data["name"],
            owner.mention,
            len(data.get("admins", [])),
            len(data.get("banned", [])),
            len(data.get("banned_chat", [])),
            len(data.get("chats", [])),
            len(data.get("subscribers", [])),
        )
        if subs := await self._get_fed_subs_str(data["_id"]):
            res += await self.text(chat.id, "fed-info-subscription") + f"{subs}"
        return res

    @command.filters(aliases=["fedadmin", "fadmin", "fadmins"])
    async def cmd_fedadmins(self, ctx: command.Context, fid: Optional[str] = None) -> str:
        """Fetch federation admins"""
        chat = ctx.chat
        user = ctx.msg.from_user

        if fid:
            data = await self.get_fed(fid)
            text = await self.text(chat.id, "fed-invalid-id")
        else:
            data = await self.get_fed_bychat(chat.id)
            text = await self.text(chat.id, "fed-no-fed-chat")

        if not data:
            return text

        if not self.is_fed_admin(data, user.id):
            return await self.text(chat.id, "fed-admin-only")

        owner = await self.bot.client.get_users(data["owner"])
        if isinstance(owner, List):
            owner = owner[0]

        text = await self.text(chat.id, "fed-admin-text", data["name"], owner.mention)
        if len(data.get("admins", [])) != 0:
            text += "\nAdmins:\n"
            admins = []
            for admin in data["admins"]:
                admins.append(admin)

            for uid in admins:
                try:
                    admin = await self.bot.client.get_users(uid)
                except PeerIdInvalid:
                    text += f"[{uid}](tg://user?id={uid})\n"
                    continue

                text += f" • {admin.mention}\n"  # type: ignore
        else:
            text += "\n" + await self.text(chat.id, "fed-no-admin")

        return text

    async def __user_fban(
        self,
        chat: Chat,
        target: User,
        banner: User,
        reason: str,
        fed_data: Mapping[str, Any],
    ) -> str:
        update = False
        if str(target.id) in fed_data.get("banned", {}).keys():
            update = True

        fullname = target.first_name + target.last_name if target.last_name else target.first_name
        await self.fban_user(fed_data["_id"], target.id, fullname=fullname, reason=reason)

        if update:
            return await self.text(
                chat.id,
                "fed-ban-info-update",
                fed_data["name"],
                banner.mention,
                target.mention,
                target.id,
                fed_data["banned"][str(target.id)]["reason"],
                reason,
            )
        return await self.text(
            chat.id,
            "fed-ban-info",
            fed_data["name"],
            banner.mention,
            target.mention,
            target.id,
            reason,
        )

    async def __channel_fban(
        self,
        chat: Chat,
        target: Chat,
        banner: User,
        reason: str,
        fed_data: Mapping[str, Any],
    ) -> str:
        update = False
        if str(target.id) in fed_data.get("banned_chat", {}).keys():
            update = True

        await self.fban_chat(fed_data["_id"], target.id, title=target.title, reason=reason)

        if update:
            return await self.text(
                chat.id,
                "fed-ban-chat-info-update",
                fed_data["name"],
                banner.mention,
                target.title,
                target.id,
                fed_data["banned_chat"][str(target.id)]["reason"],
                reason,
            )
        return await self.text(
            chat.id,
            "fed-ban-chat-info",
            fed_data["name"],
            banner.mention,
            target.title,
            target.id,
            reason,
        )

    async def _propagate_fban(
        self,
        target: Union[User, Chat],
        chats: List[int],
        host_fed: str,
        sub_fed: Optional[str] = None,
    ) -> Mapping[int, str]:
        """Propagate fban to each chat in the list"""
        failed = {}
        for chat in chats:
            try:
                await self.bot.client.ban_chat_member(chat, target.id)
            except BadRequest as br:
                self.log.warning(
                    f"Failed to send fban on subfed {sub_fed} of {host_fed} at {chat} due to {br.MESSAGE}"
                    if sub_fed
                    else f"Failed to fban {util.tg.get_username(target)} on {chat} due to {br.MESSAGE}"
                )
                failed[chat] = br.MESSAGE
            except (Forbidden, ChannelPrivate) as err:
                self.log.warning(
                    f"Can't fban on subfed {sub_fed} of {host_fed} at {chat} caused by {err.MESSAGE}"
                    if sub_fed
                    else f"Can't fban {util.tg.get_username(target)} on {chat} caused by {err.MESSAGE}"
                )
                failed[chat] = err.MESSAGE
        return failed

    async def cmd_fban(
        self, ctx: command.Context, target: Union[User, Chat, None] = None, *, reason: str = ""
    ) -> Optional[str]:
        """Fed ban command"""
        chat = ctx.chat
        if chat.type == ChatType.PRIVATE:
            return await self.text(chat.id, "err-chat-groups")

        banner = ctx.msg.from_user
        if not banner:
            return await self.text(chat.id, "err-anonymous")

        data = await self.get_fed_bychat(chat.id)
        if not data:
            return await self.text(chat.id, "fed-no-fed-chat")

        if not self.is_fed_admin(data, banner.id):
            return await self.text(chat.id, "fed-admin-only")

        reply_msg = ctx.msg.reply_to_message
        if not target:
            if ctx.args and not reply_msg:
                return await self.text(chat.id, "err-peer-invalid")

            if not reply_msg or not (reply_msg.from_user or reply_msg.sender_chat):
                return await self.text(chat.id, "fed-no-ban-user")

            target = reply_msg.from_user or reply_msg.sender_chat
            reason = ctx.input

        if target.id == self.bot.uid:
            return await self.text(chat.id, "fed-ban-self")
        if self.is_fed_admin(data, target.id):
            return await self.text(chat.id, "fed-ban-owner")
        if (
            target.id in self.bot.staff
            or target.id in (777000, 1087968824)
            or target.id == self.bot.owner
        ):
            return await self.text(chat.id, "fed-ban-protected")

        if not reason:
            reason = "No reason given."

        if isinstance(target, User):
            string = await self.__user_fban(chat, target, banner, reason, data)
        elif isinstance(target, Chat):
            string = await self.__channel_fban(chat, target, banner, reason, data)
        else:
            return await self.text(chat.id, "err-peer-invalid")

        await ctx.respond(f"Starting a federation ban for {target.id} in federation {data['name']}")
        failed = await self._propagate_fban(target, data["chats"], data["_id"])

        text = ""
        if failed:
            for key, err_msg in failed.items():
                text += f"failed to fban on chat {key} caused by {err_msg}\n\n"
                # Remove the chat federation
                await self.db.update_one({"_id": data["_id"]}, {"$pull": {"chats": chat}})
            text += f"**Those chat has leaved the federation {data['name']}!**"
            await ctx.respond(text, mode="reply", reference=ctx.response)

        if data.get("subscribers", []):
            for fed_id in data["subscribers"]:
                subs_data = await self.get_fed(fed_id)
                if not subs_data:
                    continue
                await self._propagate_fban(
                    target, subs_data.get("chats", []), data["_id"], subs_data["_id"]
                )

        await ctx.respond(string)

        # send message to federation log
        if log := data.get("log"):
            await self.bot.client.send_message(log, string, disable_web_page_preview=True)
            if failed:
                await self.bot.client.send_message(log, text)

        return None

    async def _propagate_unfban(
        self,
        target: Union[User, Chat],
        chats: List[int],
        host_fed: str,
        sub_fed: Optional[str] = None,
    ) -> None:
        for chat in chats:
            try:
                await self.bot.client.unban_chat_member(chat, target.id)
                await asyncio.sleep(0.5)
            except (BadRequest, Forbidden, ChannelPrivate) as err:
                self.log.warning(
                    f"Failed to unfban on subfed {sub_fed} of {host_fed} due to {err.MESSAGE}"
                    if sub_fed
                    else f"Failed to unfban on {host_fed} due to {err.MESSAGE}"
                )

    async def cmd_unfban(self, ctx: command.Context, target: Union[User, Chat, None] = None) -> str:
        """Unban a user on federation"""
        chat = ctx.chat
        if chat.type == ChatType.PRIVATE:
            return await self.text(chat.id, "err-chat-groups")

        banner = ctx.msg.from_user
        if not banner:
            return await self.text(chat.id, "err-anonymous")

        data = await self.get_fed_bychat(chat.id)
        if not data:
            return await self.text(chat.id, "fed-no-fed-chat")

        if not self.is_fed_admin(data, banner.id):
            return await self.text(chat.id, "fed-admin-only")

        reply_msg = ctx.msg.reply_to_message
        if not target:
            if ctx.args and not reply_msg:
                return await self.text(chat.id, "err-peer-invalid")
            if not reply_msg or not (reply_msg.from_user or reply_msg.sender_chat):
                return await self.text(chat.id, "fed-no-ban-user")
            target = reply_msg.from_user or reply_msg.sender_chat

        if (str(target.id) not in (data.get("banned", {}).keys())) and (
            str(target.id) not in (data.get("banned_chat", {}).keys())
        ):
            return await self.text(chat.id, "fed-user-not-banned")

        if isinstance(target, User):
            await self.unfban_user(data["_id"], target.id)
            text = await self.text(
                chat.id,
                "fed-unban-info",
                data["name"],
                banner.mention,
                target.mention,
                target.id,
            )
        elif isinstance(target, Chat):
            await self.unfban_chat(data["_id"], target.id)
            text = await self.text(
                chat.id,
                "fed-unban-info-chat",
                data["name"],
                banner.mention,
                target.title,
                target.id,
            )
        else:
            return ""

        await ctx.respond(f"Removing federation ban for {target.id} in federation {data['name']}")
        await self._propagate_unfban(target, data["chats"], data["_id"])

        if data.get("subscribers", []):
            for fed_id in data["subscribers"]:
                subs_data = await self.get_fed(fed_id)
                if not subs_data:
                    continue

                await self._propagate_unfban(
                    target, subs_data.get("chats", []), data["_id"], subs_data["_id"]
                )

        if log := data.get("log"):
            await self.bot.client.send_message(log, text, disable_web_page_preview=True)

        return text

    @command.filters(aliases=["fstats", "fedstats"])
    async def cmd_fbanstats(self, ctx: command.Context) -> str:
        """Get user status"""
        chat = ctx.chat
        if len(ctx.args) > 1:  # <user_id> <fed_id>
            try:
                user_id = int(ctx.args[0])
            except (TypeError, ValueError):
                return await self.text(chat.id, "fed-invalid-user-id")

            data = await self.get_fed(ctx.args[1])
            if data:
                if str(user_id) in data.get("banned", {}):
                    res = data["banned"][str(user_id)]
                    return await self.text(
                        chat.id,
                        "fed-stat-banned",
                        res["reason"],
                        res["time"].strftime("%Y %b %d %H:%M UTC"),
                    )
                if str(user_id) in data.get("banned_chat", {}):
                    res = data["banned_chat"][str(user_id)]
                    return await self.text(
                        chat.id,
                        "fed-stat-banned-chat",
                        res["reason"],
                        res["time"].strftime("%Y %b %d %H:%M UTC"),
                    )

                if str(user_id) in data.get("banned_chat", {}):
                    res = data["banned_chat"][str(user_id)]
                    return await self.text(
                        chat.id,
                        "fed-stat-banned-chat",
                        res["reason"],
                        res["time"].strftime("%Y %b %d %H:%M UTC"),
                    )
                return await self.text(chat.id, "fed-stat-not-banned")
            return await self.text(chat.id, "fed-not-found")

        user = None
        user_id = None
        if len(ctx.args) == 1:  # <user_id>
            try:
                user_id = int(ctx.args[0])
                user = await self.bot.client.get_users(user_id)
                if isinstance(user, List):
                    user = user[0]
                if not user:
                    return await self.text(chat.id, "fed-invalid-user-id")
            except (TypeError, ValueError):
                return await self.text(chat.id, "fed-invalid-user-id")

        reply_msg = ctx.msg.reply_to_message
        if not user_id:
            if reply_msg:
                user = reply_msg.from_user or reply_msg.sender_chat
            else:
                user = ctx.msg.from_user
            user_id = user.id

        if not user:
            return ""

        cursor = await self.check_fban(user_id)
        fed_list = await cursor.to_list()
        if fed_list:
            text = await self.text(chat.id, "fed-stat-multi")
            for bans in fed_list:
                text += "\n" + await self.text(
                    chat.id,
                    "fed-stat-multi-info",
                    bans["name"],
                    bans["_id"],
                    bans["banned_chat" if bans.get("banned_chat") else "banned"][str(user.id)][
                        "reason"
                    ],
                )
        else:
            text = await self.text(chat.id, "fed-stat-multi-not-banned")

        return text

    @command.filters(filters.private)
    async def cmd_fedbackup(self, ctx: command.Context) -> Optional[str]:
        """Backup federation ban list"""
        chat = ctx.chat
        user = ctx.msg.from_user

        data = await self.db.find_one({"owner": user.id})
        if not data:
            return await self.text(chat.id, "user-no-feds")

        try:
            banned = data["banned"]
        except KeyError:
            return await self.text(chat.id, "fed-backup-empty")

        file = AsyncPath(self.bot.config.DOWNLOAD_PATH + data["name"] + ".csv")

        await file.touch()
        async with file.open("w") as f:
            for banned_user in banned:
                ban_data = banned[banned_user]
                await f.write(
                    f"{banned_user},{ban_data['name']},{ban_data['reason']},{ban_data['time']}\n"
                )

        await ctx.respond(document=str(file))
        await file.unlink()
        return None

    @command.filters(filters.private)
    async def cmd_fedrestore(self, ctx: command.Context) -> Optional[str]:
        """Restore a backup bans"""
        chat = ctx.chat
        user = ctx.msg.from_user
        reply_msg = ctx.msg.reply_to_message
        if not (reply_msg and reply_msg.document):
            return await self.text(chat.id, "no-backup-file")

        data = await self.db.find_one({"owner": user.id})
        if not data:
            return await self.text(chat.id, "user-no-feds")

        file = AsyncPath(await reply_msg.download(self.bot.config.DOWNLOAD_PATH))

        fid = data["_id"]
        tasks = []  # type: List[asyncio.Task[None]]
        async with file.open("r") as f:
            line: str
            async for line in f:  # type: ignore
                d = line.split(",")
                task = self.bot.loop.create_task(
                    self.fban_user(fid, int(d[0]), fullname=d[1], reason=d[2])
                )
                tasks.append(task)

        ret = await asyncio.gather(self.text(chat.id, "fed-restore-done"), file.unlink(), *tasks)
        return ret[0]

    @command.filters(filters.private, aliases=["myfeds"])
    async def cmd_myfed(self, ctx: command.Context) -> str:
        """Get current users federation"""
        chat = ctx.chat
        user = ctx.msg.from_user

        data = await self.db.find_one({"owner": user.id})
        if data:
            return (
                await self.text(chat.id, "fed-myfeds-owner")
                + f"\n- `{data['_id']}`: {data['name']}"
            )

        cursor = self.db.find({"admins": user.id}, {"_id": 1, "name": 1})
        fed_list = await cursor.to_list()
        if fed_list:
            text = await self.text(chat.id, "fed-myfeds-admin") + "\n"
            for fed in fed_list:
                text += f"- `{fed['_id']}`: {fed['name']}\n"

            return text

        return await self.text(chat.id, "fed-myfeds-no-admin")

    def generate_log_btn(self, data: Mapping[str, Any]) -> InlineKeyboardMarkup:
        """Generate log channel verify button"""
        return InlineKeyboardMarkup(
            [
                [
                    InlineKeyboardButton(
                        text="Click to confirm identity",
                        callback_data=f"logfed_{data['owner']}_{data['_id']}",
                    )
                ],
            ]
        )

    async def channel_setlog(self, message: Message):
        fid = message.text.split(" ")
        if not fid or len(fid) < 2:
            await message.reply_text(await self.text(message.chat.id, "fed-set-log-args"))
            return
        data = await self.get_fed(fid[1])
        if not data:
            await message.reply_text(await self.text(message.chat.id, "fed-not-found"))
            return
        await message.reply_text(
            await self.text(message.chat.id, "fed-check-identity"),
            reply_markup=self.generate_log_btn(data),
        )

    async def cmd_setfedlog(self, ctx: command.Context, fid: Optional[str] = None) -> Optional[str]:
        """Set a federation log channel"""
        chat = ctx.chat
        if chat.type == ChatType.CHANNEL:
            if not fid:
                return await self.text(chat.id, "fed-set-log-args")

            data = await self.get_fed(fid)
            if not data:
                return await self.text(chat.id, "Fed not found!")

            await ctx.respond(
                await self.text(chat.id, "fed-check-identity"),
                reply_markup=InlineKeyboardMarkup(
                    [
                        [
                            InlineKeyboardButton(
                                text="Click to confirm identity",
                                callback_data=f"logfed_{data['owner']}_{data['_id']}",
                            )
                        ],
                    ]
                ),
            )
            return None

        if chat.type in {ChatType.GROUP, ChatType.SUPERGROUP}:
            data = await self.get_fed_byowner(ctx.author.id)
            if not data:
                return await self.text(chat.id, "user-no-feds")

            ret, _ = await asyncio.gather(
                self.text(chat.id, "fed-log-set-group", name=data["name"]),
                self.db.update_one({"_id": data["_id"]}, {"$set": {"log": chat.id}}),
            )
            return ret

        return await self.text(chat.id, "err-chat-groups")

    async def cmd_unsetfedlog(self, ctx: command.Context) -> str:
        """Unset the federation log channel"""
        chat = ctx.chat
        user = ctx.msg.from_user
        if chat.type == ChatType.PRIVATE:
            data = await self.get_fed_byowner(user.id)
            if not data:
                return await self.text(chat.id, "user-no-feds")

            ret, _ = await asyncio.gather(
                self.text(chat.id, "fed-log-unset", data["name"]),
                self.db.update_one({"_id": data["_id"]}, {"$set": {"log": None}}),
            )
            return ret

        return await self.text(chat.id, "err-chat-private")

    async def _subfed_perm_check(self, ctx: command.Context, fid: Optional[str] = None):
        """Check if able to subscribe to a fed and returns current and target federation"""
        if not fid:
            await ctx.respond(await self.text(ctx.chat.id, "fed-not-found"))
            return None
        if ctx.chat.type == ChatType.PRIVATE:
            await ctx.respond(await self.text(ctx.chat.id, "err-chat-groups"))
            return None

        curr_fed = await self.get_fed_bychat(ctx.chat.id)
        target = await self.get_fed(fid)
        if not curr_fed:
            await ctx.respond(await self.text(ctx.chat.id, "fed-no-fed-chat"))
            return None
        if curr_fed["owner"] != ctx.author.id:
            await ctx.respond(await self.text(ctx.chat.id, "fed-owner-cmd"))
            return None
        if not target:
            await ctx.respond(await self.text(ctx.chat.id, "fed-not-found"))
            return None

        return (curr_fed, target)

    async def cmd_subfed(self, ctx: command.Context, fid: Optional[str] = None):
        """Subscribe to a federation"""
        res = await self._subfed_perm_check(ctx, fid)
        if not res:
            return
        curr_fed, to_subs = res

        await self.db.update_one({"_id": fid}, {"$push": {"subscribers": curr_fed["_id"]}})
        try:
            await self.bot.client.send_message(
                to_subs["log"] or to_subs["owner"],
                f"Federation {curr_fed['name']} has subscribed to {to_subs['name']}",
            )
        except PeerIdInvalid:
            self.log.warning("Failed to send fed-subs message to log channel")

        return await self.text(ctx.chat.id, "fed-subs-join", curr_fed["name"], to_subs["name"])

    async def cmd_unsubfed(self, ctx: command.Context, fid: Optional[str] = None):
        """Unsubscribe from a federation"""
        res = await self._subfed_perm_check(ctx, fid)
        if not res:
            return
        curr_fed, to_unsubs = res

        await self.db.update_one({"_id": fid}, {"$pull": {"subscribers": curr_fed["_id"]}})
        try:
            await self.bot.client.send_message(
                to_unsubs["log"],
                f"Federation **{curr_fed['name']}** has unsubscribed from **{to_unsubs['name']}**",
            )
        except PeerIdInvalid:
            self.log.warning("Failed to send unsubs message to log channel")

        return await self.text(ctx.chat.id, "fed-subs-leave", curr_fed["name"], to_unsubs["name"])



"""Plugin Backup and Restore"""
# Copyright (C) 2020 - 2023  UserbotIndo Team, <https://github.com/userbotindo.git>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import asyncio
import json
from datetime import datetime
from typing import Any, ClassVar, MutableMapping, Optional

from aiopath import AsyncPath

from anjani import command, filters, plugin


class Backups(plugin.Plugin):
    name: ClassVar[str] = "Backups"
    helpable: ClassVar[bool] = True

    async def on_load(self) -> None:
        self.db = self.bot.db.get_collection("MIGRATED_BACKUPS")

    @command.filters(filters.admin_only)
    async def cmd_backup(self, ctx: command.Context) -> Optional[str]:
        """Backup chat data from file"""
        chat = ctx.chat
        data = {"chat_id": chat.id}
        file = AsyncPath(f"{chat.title}-backup.anjani")

        await ctx.respond(await self.text(chat.id, "backup-progress"))

        results = await self.bot.dispatch_event("plugin_backup", chat.id)
        if not results or len(results) <= 1:
            return await self.text(chat.id, "backup-null")

        result: MutableMapping[str, Any]
        for result in results:
            data.update(result)

        await file.write_text(json.dumps(data, indent=2))

        saved = ""
        del data["chat_id"]
        for key in data:
            saved += f"\n× `{key}`"

        date = datetime.now().strftime("%H:%M - %d/%b/%Y")
        await asyncio.gather(
            ctx.msg.reply_document(
                str(file),
                caption=await self.text(chat.id, "backup-doc", chat.title, chat.id, date, saved),
            ),
            ctx.response.delete(),
        )
        await file.unlink()

        return None

    @command.filters(filters.admin_only)
    async def cmd_restore(self, ctx: command.Context) -> Optional[str]:
        """Restore data to a file"""
        chat = ctx.chat
        reply_msg = ctx.msg.reply_to_message

        if not reply_msg or (reply_msg and not reply_msg.document):
            return await self.text(chat.id, "no-backup-file")

        await ctx.respond(await self.text(chat.id, "restore-progress"))

        file = AsyncPath(await ctx.msg.reply_to_message.download())

        try:
            data: MutableMapping[str, Any] = json.loads(await file.read_text())
        except json.JSONDecodeError:
            return await self.text(chat.id, "invalid-backup-file")

        try:  # also check if the file isn't a valid backup file
            if data["chat_id"] != chat.id:
                return await self.text(chat.id, "backup-id-invalid")
        except KeyError:
            return await self.text(chat.id, "invalid-backup-file")

        if len(data) <= 1:
            return await self.text(chat.id, "backup-data-null")

        if data.get("_migrated", False):
            await ctx.respond(
                await self.text(chat.id, "restore-progress") + "\nMigrate file detected..."
            )
            await self.db.insert_one({"chat_id": chat.id, "on": datetime.now()})

        await self.bot.dispatch_event("plugin_restore", chat.id, data)
        await asyncio.gather(ctx.respond(await self.text(chat.id, "backup-done")), file.unlink())
        return None



"""Bot rules command"""
# Copyright (C) 2020 - 2023  UserbotIndo Team, <https://github.com/userbotindo.git>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import asyncio
from typing import Any, MutableMapping, Optional

from pyrogram.errors import PeerIdInvalid
from pyrogram.types import InlineKeyboardButton, InlineKeyboardMarkup, Message

from anjani import command, filters, plugin, util


class Rules(plugin.Plugin):
    name = "Rules"
    helpable = True

    db: util.db.AsyncCollection

    async def on_load(self) -> None:
        self.db = self.bot.db.get_collection("RULES")

    async def on_chat_migrate(self, message: Message) -> None:
        new_chat = message.chat.id
        old_chat = message.migrate_from_chat_id

        await self.db.update_one(
            {"chat_id": old_chat},
            {"$set": {"chat_id": new_chat}},
        )

    async def on_plugin_backup(self, chat_id: int) -> MutableMapping[str, Any]:
        rules = await self.db.find_one({"chat_id": chat_id}, {"_id": False})
        return {self.name: rules} if rules else {}

    async def on_plugin_restore(self, chat_id: int, data: MutableMapping[str, Any]) -> None:
        await self.db.update_one({"chat_id": chat_id}, {"$set": data[self.name]}, upsert=True)

    @command.filters(filters.admin_only)
    async def cmd_setrules(self, ctx: command.Context) -> str:
        chat = ctx.chat
        if not ctx.input_raw:
            return await self.text(chat.id, "rules-blank-err")

        content = ctx.input_raw
        ret, _ = await asyncio.gather(
            self.text(chat.id, "rules-set", f"t.me/{self.bot.user.username}?start=rules_{chat.id}"),
            self.db.update_one({"chat_id": chat.id}, {"$set": {"rules": content}}, upsert=True),
        )
        return ret

    @command.filters(filters.admin_only)
    async def cmd_clearrules(self, ctx: command.Context) -> str:
        chat = ctx.chat
        ret, _ = await asyncio.gather(
            self.text(chat.id, "rules-clear"), self.db.delete_one({"chat_id": chat.id})
        )
        return ret

    async def cmd_rules(self, ctx: command.Context) -> Optional[str]:
        chat = ctx.chat
        content = await self.db.find_one({"chat_id": chat.id})
        if not content:
            return await self.text(chat.id, "rules-none")

        await ctx.respond(
            await self.text(chat.id, "rules-view-caption"),
            reply_markup=InlineKeyboardMarkup(
                [
                    [
                        InlineKeyboardButton(
                            text=await self.text(chat.id, "rules-button"),
                            url=f"t.me/{self.bot.user.username}?start=rules_{chat.id}",
                        )
                    ]
                ]
            ),
        )
        return None

    async def start_rules(self, ctx: command.Context) -> str:
        try:
            rules_id = int(ctx.input.split("_")[1])
        except ValueError:
            return await self.text(ctx.chat.id, "rules-invalid-button", ctx.input.split("_")[1])
        try:
            content, chat = await asyncio.gather(
                self.db.find_one({"chat_id": rules_id}),
                self.bot.client.get_chat(rules_id),
            )
        except PeerIdInvalid:
            content, chat = None, None

        if not content or not chat:
            return await self.text(rules_id, "rules-none")

        text = await self.text(rules_id, "rules-view-pm", chat.title)

        return text + content["rules"]
